# pylint: disable=invalid-overridden-method
import time
from test.testbase import DbIsolatedTest, run_as_sync

from api.routes.analysis import AnalysisUpdateModel, update_analysis
from db.python.filters import GenericFilter
from db.python.layers.analysis import AnalysisLayer
from db.python.layers.assay import AssayLayer
from db.python.layers.participant import ParticipantLayer
from db.python.layers.sample import SampleLayer
from db.python.layers.sequencing_group import SequencingGroupLayer
from db.python.tables.analysis import AnalysisFilter
from models.enums import AnalysisStatus
from models.models import (
    PRIMARY_EXTERNAL_ORG,
    AnalysisInternal,
    AssayUpsertInternal,
    ParticipantUpsertInternal,
    SampleUpsertInternal,
    SequencingGroupUpsertInternal,
    parse_sql_bool,
)


class TestAnalysis(DbIsolatedTest):
    """Test sample class"""

    # pylint: disable=too-many-instance-attributes

    @run_as_sync
    async def setUp(self) -> None:
        # don't need to await because it's tagged @run_as_sync
        super().setUp()
        self.sl = SampleLayer(self.connection)
        self.sgl = SequencingGroupLayer(self.connection)
        self.asl = AssayLayer(self.connection)
        self.al = AnalysisLayer(self.connection)
        self.pl = ParticipantLayer(self.connection)

        sample = await self.sl.upsert_sample(
            SampleUpsertInternal(
                external_ids={PRIMARY_EXTERNAL_ORG: 'Test01'},
                type='blood',
                meta={'meta': 'meta ;)'},
                active=True,
                sequencing_groups=[
                    SequencingGroupUpsertInternal(
                        type='genome',
                        technology='short-read',
                        platform='illumina',
                        meta={},
                        sample_id=None,
                        assays=[
                            AssayUpsertInternal(
                                type='sequencing',
                                meta={
                                    'sequencing_type': 'genome',
                                    'sequencing_technology': 'short-read',
                                    'sequencing_platform': 'illumina',
                                },
                            )
                        ],
                    ),
                    SequencingGroupUpsertInternal(
                        type='exome',
                        technology='short-read',
                        platform='illumina',
                        meta={},
                        sample_id=None,
                        assays=[
                            AssayUpsertInternal(
                                type='sequencing',
                                meta={
                                    'sequencing_type': 'exome',
                                    'sequencing_technology': 'short-read',
                                    'sequencing_platform': 'illumina',
                                },
                            )
                        ],
                    ),
                ],
            )
        )
        self.sample_id = sample.id
        self.genome_sequencing_group_id = sample.sequencing_groups[0].id
        self.exome_sequencing_group_id = sample.sequencing_groups[self.project_id].id

    @run_as_sync
    async def test_get_analysis_by_id(self):
        """
        Test getting an analysis by id
        """
        analysis_id = await self.al.create_analysis(
            AnalysisInternal(
                type='cram',
                status=AnalysisStatus.COMPLETED,
                sequencing_group_ids=[self.genome_sequencing_group_id],
                meta={'sequencing_type': 'genome', 'size': 1024},
            )
        )

        analysis = await self.al.get_analysis_by_id(analysis_id)
        self.assertEqual(analysis_id, analysis.id)
        self.assertEqual('cram', analysis.type)
        self.assertEqual(AnalysisStatus.COMPLETED, analysis.status)

    @run_as_sync
    async def test_empty_query(self):
        """
        Test empty IDs to see the query construction
        """
        analyses = await self.al.query(AnalysisFilter(id=GenericFilter(in_=[])))
        self.assertEqual(len(analyses), 0)

    @run_as_sync
    async def test_add_cram(self):
        """
        Test adding an analysis of type CRAM
        """

        analysis_id = await self.al.create_analysis(
            AnalysisInternal(
                type='cram',
                status=AnalysisStatus.COMPLETED,
                sequencing_group_ids=[self.genome_sequencing_group_id],
                meta={'sequencing_type': 'genome', 'size': 1024},
            )
        )

        analyses = await self.connection.connection.fetch_all('SELECT * FROM analysis')
        analysis_sgs = await self.connection.connection.fetch_all(
            'SELECT * FROM analysis_sequencing_group'
        )

        self.assertEqual(1, len(analyses))
        self.assertEqual(analysis_id, analyses[0]['id'])
        self.assertEqual(1, analysis_sgs[0]['sequencing_group_id'])
        self.assertEqual(analyses[0]['id'], analysis_sgs[0]['analysis_id'])

    @run_as_sync
    async def test_get_analysis(self):
        """
        Test adding an analysis of type ANALYSIS_RUNNER
        """

        a_id = await self.al.create_analysis(
            AnalysisInternal(
                type='analysis-runner',
                status=AnalysisStatus.UNKNOWN,
                sequencing_group_ids=[],
                meta={},
            )
        )

        analyses = await self.al.query(
            AnalysisFilter(
                project=GenericFilter(eq=self.project_id),
                type=GenericFilter(eq='analysis-runner'),
            )
        )
        expected = [
            AnalysisInternal(
                id=a_id,
                type='analysis-runner',
                status=AnalysisStatus.UNKNOWN,
                sequencing_group_ids=[],
                cohort_ids=[],
                output=None,
                timestamp_completed=None,
                project=1,
                meta={},
                active=True,
                author=None,
            )
        ]

        self.assertEqual(analyses, expected)

    @run_as_sync
    async def test_get_sample_cram_path_map_for_seqr(self):
        """
        Exercise get_sample_cram_path_map_for_seqr()
        """

        part = await self.pl.upsert_participants(
            [
                ParticipantUpsertInternal(
                    external_ids={PRIMARY_EXTERNAL_ORG: 'PEXT1'},
                    meta={},
                    samples=[SampleUpsertInternal(id=self.sample_id)],
                ),
            ],
        )

        id_map = await self.al.get_sample_cram_path_map_for_seqr(
            self.project_id, ['blood'], [part[0].id]
        )
        self.assertIsInstance(id_map, list)

    @run_as_sync
    async def test_get_sgs_by_analysis_id_with_no_eids(self):
        """
        Test get_sgs_by_analysis_id()
        """

        # duplicate here to ensure sgs don't have any external ids
        assay_meta = {
            'sequencing_type': 'genome',
            'sequencing_technology': 'short-read',
            'sequencing_platform': 'illumina',
        }
        sample = await self.sl.upsert_sample(
            SampleUpsertInternal(
                external_ids={PRIMARY_EXTERNAL_ORG: 'test_sgs_aid'},
                type='blood',
                meta={},
                active=True,
                sequencing_groups=[
                    SequencingGroupUpsertInternal(
                        type='genome',
                        technology='short-read',
                        platform='illumina',
                        assays=[
                            AssayUpsertInternal(
                                type='sequencing',
                                meta=assay_meta,
                            )
                        ],
                    ),
                    SequencingGroupUpsertInternal(
                        type='exome',
                        technology='short-read',
                        platform='illumina',
                        assays=[
                            AssayUpsertInternal(
                                type='sequencing',
                                meta=assay_meta,
                            )
                        ],
                    ),
                ],
            )
        )

        genome_id = sample.sequencing_groups[0].id
        exome_id = sample.sequencing_groups[1].id

        a_id = await self.al.create_analysis(
            AnalysisInternal(
                type='analysis-runner',
                status=AnalysisStatus.UNKNOWN,
                sequencing_group_ids=[genome_id, exome_id],
                meta={},
            )
        )

        sgs_by_aid = await self.sgl.get_sequencing_groups_by_analysis_ids([a_id])
        self.assertIn(a_id, sgs_by_aid)
        sgs = sorted(sgs_by_aid[a_id], key=lambda sg: sg.id)

        self.assertEqual(sgs[0].id, genome_id)
        self.assertEqual(sgs[1].id, exome_id)

    @run_as_sync
    async def test_update_analysis(self):
        """
        Test Analysis update
        """

        # create an analysis
        a_id = await self.al.create_analysis(
            AnalysisInternal(
                type='analysis-runner',
                status=AnalysisStatus.COMPLETED,
                sequencing_group_ids=[],
                meta={},
            ),
        )

        # get the timestamp_completed of the analysis
        init_analyses = await self.al.query(
            AnalysisFilter(
                project=GenericFilter(eq=self.project_id),
                type=GenericFilter(eq='analysis-runner'),
            )
        )

        # store the timestamp_completed
        init_timestamp_completed = init_analyses[0].timestamp_completed

        # be sure the now is different than before
        time.sleep(2)

        # update the analysis with some new data
        await self.al.update_analysis(
            a_id,
            status=AnalysisStatus.COMPLETED,
            meta={'sequencing_type': 'genome', 'size': 1024},
            output='test_output',
        )

        # check the analysis after update
        # be sure timestamp_completed has not been touched
        analyses = await self.al.query(
            AnalysisFilter(
                project=GenericFilter(eq=self.project_id),
                type=GenericFilter(eq='analysis-runner'),
            ),
        )
        expected = [
            AnalysisInternal(
                id=a_id,
                type='analysis-runner',
                status=AnalysisStatus.COMPLETED,
                sequencing_group_ids=[],
                cohort_ids=[],
                output='test_output',
                outputs='test_output',
                timestamp_completed=init_timestamp_completed,
                project=1,
                meta={'sequencing_type': 'genome', 'size': 1024},
                active=True,
                author=None,
            ),
        ]

        self.assertEqual(analyses, expected)

    @run_as_sync
    async def test_route_update_active(self):
        """
        Test that update_analysis(active=False) is effective
        """
        analysis_id = await self.al.create_analysis(
            AnalysisInternal(
                type='cram',
                status=AnalysisStatus.COMPLETED,
                sequencing_group_ids=[self.genome_sequencing_group_id],
            )
        )

        analyses = await self.connection.connection.fetch_all('SELECT * FROM analysis')
        self.assertEqual(1, len(analyses))
        self.assertEqual(analysis_id, analyses[0]['id'])
        self.assertTrue(parse_sql_bool(analyses[0]['active']))

        inactivate = AnalysisUpdateModel(active=False, status=AnalysisStatus.COMPLETED)
        await update_analysis(analysis_id, inactivate, self.connection)

        analyses = await self.connection.connection.fetch_all('SELECT * FROM analysis')
        self.assertEqual(1, len(analyses))
        self.assertEqual(analysis_id, analyses[0]['id'])
        self.assertFalse(parse_sql_bool(analyses[0]['active']))

        analyses = await self.al.query(
            AnalysisFilter(project=GenericFilter(eq=self.project_id))
        )
        self.assertEqual(1, len(analyses))
        self.assertEqual(analysis_id, analyses[0].id)
        self.assertFalse(analyses[0].active)



import unittest
from typing import Any
from unittest import mock

import google.cloud.bigquery as bq

from db.python.gcp_connect import BqConnection
from db.python.layers.billing import BillingLayer


class BqTest(unittest.TestCase):
    """Base class for Big Query integration tests"""

    # author and grp_project are not used in the BQ tests, but are required
    # so some dummy values are preset
    author: str = 'Author'
    gcp_project: str = 'GCP_PROJECT'

    bq_result: bq.job.QueryJob
    bq_client: bq.Client
    connection: BqConnection
    table_obj: Any | None = None

    def setUp(self) -> None:
        super().setUp()

        # Mockup BQ results
        self.bq_result = mock.MagicMock(spec=bq.job.QueryJob)

        # mock BigQuery client
        self.bq_client = mock.MagicMock(spec=bq.Client)
        self.bq_client.query.return_value = self.bq_result

        # Mock BqConnection
        self.connection = mock.MagicMock(spec=BqConnection)
        self.connection.gcp_project = self.gcp_project
        self.connection.connection = self.bq_client
        self.connection.author = self.author

        # Mockup BillingLayer
        self.layer = BillingLayer(self.connection)

        # overwrite table object in inhereted tests:
        self.table_obj = None



"""
    Sample metadata API

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: 7.4.1
    Generated by: https://openapi-generator.tech
"""


import re  # noqa: F401
import sys  # noqa: F401

from metamist.api_client import ApiClient, Endpoint as _Endpoint
from metamist.model_utils import (  # noqa: F401
    check_allowed_values,
    check_validations,
    date,
    datetime,
    file_type,
    none_type,
    validate_and_convert_types,
    async_wrap
)
from metamist.model.http_validation_error import HTTPValidationError
from metamist.model.project import Project
from metamist.model.project_member_update import ProjectMemberUpdate


class ProjectApi(object):
    """NOTE: This class is auto generated by OpenAPI Generator
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    def __init__(self, api_client=None):
        if api_client is None:
            api_client = ApiClient()
        self.api_client = api_client

        self.create_project_async = async_wrap(self.create_project)
        self.create_project_endpoint = _Endpoint(
            settings={
                'response_type': (int,),
                'auth': [
                    'HTTPBearer'
                ],
                'endpoint_path': '/api/v1/project/',
                'operation_id': 'create_project',
                'http_method': 'PUT',
                'servers': None,
            },
            params_map={
                'all': [
                    'name',
                    'dataset',
                    'create_test_project',
                ],
                'required': [
                    'name',
                    'dataset',
                ],
                'nullable': [
                ],
                'enum': [
                ],
                'validation': [
                ]
            },
            root_map={
                'validations': {
                },
                'allowed_values': {
                },
                'openapi_types': {
                    'name':
                        (str,),
                    'dataset':
                        (str,),
                    'create_test_project':
                        (bool,),
                },
                'attribute_map': {
                    'name': 'name',
                    'dataset': 'dataset',
                    'create_test_project': 'create_test_project',
                },
                'location_map': {
                    'name': 'query',
                    'dataset': 'query',
                    'create_test_project': 'query',
                },
                'collection_format_map': {
                }
            },
            headers_map={
                'accept': [
                    'application/json'
                ],
                'content_type': [],
            },
            api_client=api_client
        )

        self.delete_project_data_async = async_wrap(self.delete_project_data)
        self.delete_project_data_endpoint = _Endpoint(
            settings={
                'response_type': (bool, date, datetime, dict, float, int, list, str, none_type,),
                'auth': [
                    'HTTPBearer'
                ],
                'endpoint_path': '/api/v1/project/{project}',
                'operation_id': 'delete_project_data',
                'http_method': 'DELETE',
                'servers': None,
            },
            params_map={
                'all': [
                    'project',
                    'delete_project',
                ],
                'required': [
                    'project',
                ],
                'nullable': [
                ],
                'enum': [
                ],
                'validation': [
                ]
            },
            root_map={
                'validations': {
                },
                'allowed_values': {
                },
                'openapi_types': {
                    'project':
                        (str,),
                    'delete_project':
                        (bool,),
                },
                'attribute_map': {
                    'project': 'project',
                    'delete_project': 'delete_project',
                },
                'location_map': {
                    'project': 'path',
                    'delete_project': 'query',
                },
                'collection_format_map': {
                }
            },
            headers_map={
                'accept': [
                    'application/json'
                ],
                'content_type': [],
            },
            api_client=api_client
        )

        self.get_all_projects_async = async_wrap(self.get_all_projects)
        self.get_all_projects_endpoint = _Endpoint(
            settings={
                'response_type': ([Project],),
                'auth': [
                    'HTTPBearer'
                ],
                'endpoint_path': '/api/v1/project/all',
                'operation_id': 'get_all_projects',
                'http_method': 'GET',
                'servers': None,
            },
            params_map={
                'all': [
                ],
                'required': [],
                'nullable': [
                ],
                'enum': [
                ],
                'validation': [
                ]
            },
            root_map={
                'validations': {
                },
                'allowed_values': {
                },
                'openapi_types': {
                },
                'attribute_map': {
                },
                'location_map': {
                },
                'collection_format_map': {
                }
            },
            headers_map={
                'accept': [
                    'application/json'
                ],
                'content_type': [],
            },
            api_client=api_client
        )

        self.get_my_projects_async = async_wrap(self.get_my_projects)
        self.get_my_projects_endpoint = _Endpoint(
            settings={
                'response_type': ([str],),
                'auth': [
                    'HTTPBearer'
                ],
                'endpoint_path': '/api/v1/project/',
                'operation_id': 'get_my_projects',
                'http_method': 'GET',
                'servers': None,
            },
            params_map={
                'all': [
                ],
                'required': [],
                'nullable': [
                ],
                'enum': [
                ],
                'validation': [
                ]
            },
            root_map={
                'validations': {
                },
                'allowed_values': {
                },
                'openapi_types': {
                },
                'attribute_map': {
                },
                'location_map': {
                },
                'collection_format_map': {
                }
            },
            headers_map={
                'accept': [
                    'application/json'
                ],
                'content_type': [],
            },
            api_client=api_client
        )

        self.get_seqr_projects_async = async_wrap(self.get_seqr_projects)
        self.get_seqr_projects_endpoint = _Endpoint(
            settings={
                'response_type': (bool, date, datetime, dict, float, int, list, str, none_type,),
                'auth': [
                    'HTTPBearer'
                ],
                'endpoint_path': '/api/v1/project/seqr/all',
                'operation_id': 'get_seqr_projects',
                'http_method': 'GET',
                'servers': None,
            },
            params_map={
                'all': [
                ],
                'required': [],
                'nullable': [
                ],
                'enum': [
                ],
                'validation': [
                ]
            },
            root_map={
                'validations': {
                },
                'allowed_values': {
                },
                'openapi_types': {
                },
                'attribute_map': {
                },
                'location_map': {
                },
                'collection_format_map': {
                }
            },
            headers_map={
                'accept': [
                    'application/json'
                ],
                'content_type': [],
            },
            api_client=api_client
        )

        self.update_project_async = async_wrap(self.update_project)
        self.update_project_endpoint = _Endpoint(
            settings={
                'response_type': (bool, date, datetime, dict, float, int, list, str, none_type,),
                'auth': [
                    'HTTPBearer'
                ],
                'endpoint_path': '/api/v1/project/{project}/update',
                'operation_id': 'update_project',
                'http_method': 'POST',
                'servers': None,
            },
            params_map={
                'all': [
                    'project',
                    'body',
                ],
                'required': [
                    'project',
                    'body',
                ],
                'nullable': [
                ],
                'enum': [
                ],
                'validation': [
                ]
            },
            root_map={
                'validations': {
                },
                'allowed_values': {
                },
                'openapi_types': {
                    'project':
                        (str,),
                    'body':
                        ({str: (bool, date, datetime, dict, float, int, list, str, none_type)},),
                },
                'attribute_map': {
                    'project': 'project',
                },
                'location_map': {
                    'project': 'path',
                    'body': 'body',
                },
                'collection_format_map': {
                }
            },
            headers_map={
                'accept': [
                    'application/json'
                ],
                'content_type': [
                    'application/json'
                ]
            },
            api_client=api_client
        )

        self.update_project_members_async = async_wrap(self.update_project_members)
        self.update_project_members_endpoint = _Endpoint(
            settings={
                'response_type': (bool, date, datetime, dict, float, int, list, str, none_type,),
                'auth': [
                    'HTTPBearer'
                ],
                'endpoint_path': '/api/v1/project/{project}/members',
                'operation_id': 'update_project_members',
                'http_method': 'PATCH',
                'servers': None,
            },
            params_map={
                'all': [
                    'project',
                    'project_member_update',
                ],
                'required': [
                    'project',
                    'project_member_update',
                ],
                'nullable': [
                ],
                'enum': [
                ],
                'validation': [
                ]
            },
            root_map={
                'validations': {
                },
                'allowed_values': {
                },
                'openapi_types': {
                    'project':
                        (str,),
                    'project_member_update':
                        ([ProjectMemberUpdate],),
                },
                'attribute_map': {
                    'project': 'project',
                },
                'location_map': {
                    'project': 'path',
                    'project_member_update': 'body',
                },
                'collection_format_map': {
                }
            },
            headers_map={
                'accept': [
                    'application/json'
                ],
                'content_type': [
                    'application/json'
                ]
            },
            api_client=api_client
        )

    def create_project(
        self,
        name,
        dataset,
        **kwargs
    ):
        """Create Project  # noqa: E501

        Create a new project  # noqa: E501
        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please pass async_req=True

        >>> thread = api.create_project(name, dataset, async_req=True)
        >>> result = thread.get()

        Args:
            name (str):
            dataset (str):

        Keyword Args:
            create_test_project (bool): [optional] if omitted the server will use the default value of False
            _return_http_data_only (bool): response data without head status
                code and headers. Default is True.
            _preload_content (bool): if False, the urllib3.HTTPResponse object
                will be returned without reading/decoding response data.
                Default is True.
            _request_timeout (int/float/tuple): timeout setting for this request. If
                one number provided, it will be total request timeout. It can also
                be a pair (tuple) of (connection, read) timeouts.
                Default is None.
            _check_input_type (bool): specifies if type checking
                should be done one the data sent to the server.
                Default is True.
            _check_return_type (bool): specifies if type checking
                should be done one the data received from the server.
                Default is True.
            _content_type (str/None): force body content-type.
                Default is None and content-type will be predicted by allowed
                content-types and body.
            _host_index (int/None): specifies the index of the server
                that we want to use.
                Default is read from the configuration.
            async_req (bool): execute request asynchronously

        Returns:
            int
                If the method is called asynchronously, returns the request
                thread.
        """
        kwargs['async_req'] = kwargs.get(
            'async_req', False
        )
        kwargs['_return_http_data_only'] = kwargs.get(
            '_return_http_data_only', True
        )
        kwargs['_preload_content'] = kwargs.get(
            '_preload_content', True
        )
        kwargs['_request_timeout'] = kwargs.get(
            '_request_timeout', None
        )
        kwargs['_check_input_type'] = kwargs.get(
            '_check_input_type', True
        )
        kwargs['_check_return_type'] = kwargs.get(
            '_check_return_type', True
        )
        kwargs['_content_type'] = kwargs.get(
            '_content_type')
        kwargs['_host_index'] = kwargs.get('_host_index')
        kwargs['name'] = \
            name
        kwargs['dataset'] = \
            dataset
        return self.create_project_endpoint.call_with_http_info(**kwargs)

    def delete_project_data(
        self,
        project,
        **kwargs
    ):
        """Delete Project Data  # noqa: E501

        Delete all data in a project by project name. Can optionally delete the project itself. Requires READ access + project-creator permissions  # noqa: E501
        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please pass async_req=True

        >>> thread = api.delete_project_data(project, async_req=True)
        >>> result = thread.get()

        Args:
            project (str):

        Keyword Args:
            delete_project (bool): [optional] if omitted the server will use the default value of False
            _return_http_data_only (bool): response data without head status
                code and headers. Default is True.
            _preload_content (bool): if False, the urllib3.HTTPResponse object
                will be returned without reading/decoding response data.
                Default is True.
            _request_timeout (int/float/tuple): timeout setting for this request. If
                one number provided, it will be total request timeout. It can also
                be a pair (tuple) of (connection, read) timeouts.
                Default is None.
            _check_input_type (bool): specifies if type checking
                should be done one the data sent to the server.
                Default is True.
            _check_return_type (bool): specifies if type checking
                should be done one the data received from the server.
                Default is True.
            _content_type (str/None): force body content-type.
                Default is None and content-type will be predicted by allowed
                content-types and body.
            _host_index (int/None): specifies the index of the server
                that we want to use.
                Default is read from the configuration.
            async_req (bool): execute request asynchronously

        Returns:
            bool, date, datetime, dict, float, int, list, str, none_type
                If the method is called asynchronously, returns the request
                thread.
        """
        kwargs['async_req'] = kwargs.get(
            'async_req', False
        )
        kwargs['_return_http_data_only'] = kwargs.get(
            '_return_http_data_only', True
        )
        kwargs['_preload_content'] = kwargs.get(
            '_preload_content', True
        )
        kwargs['_request_timeout'] = kwargs.get(
            '_request_timeout', None
        )
        kwargs['_check_input_type'] = kwargs.get(
            '_check_input_type', True
        )
        kwargs['_check_return_type'] = kwargs.get(
            '_check_return_type', True
        )
        kwargs['_content_type'] = kwargs.get(
            '_content_type')
        kwargs['_host_index'] = kwargs.get('_host_index')
        kwargs['project'] = \
            project
        return self.delete_project_data_endpoint.call_with_http_info(**kwargs)

    def get_all_projects(
        self,
        **kwargs
    ):
        """Get All Projects  # noqa: E501

        Get list of projects  # noqa: E501
        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please pass async_req=True

        >>> thread = api.get_all_projects(async_req=True)
        >>> result = thread.get()


        Keyword Args:
            _return_http_data_only (bool): response data without head status
                code and headers. Default is True.
            _preload_content (bool): if False, the urllib3.HTTPResponse object
                will be returned without reading/decoding response data.
                Default is True.
            _request_timeout (int/float/tuple): timeout setting for this request. If
                one number provided, it will be total request timeout. It can also
                be a pair (tuple) of (connection, read) timeouts.
                Default is None.
            _check_input_type (bool): specifies if type checking
                should be done one the data sent to the server.
                Default is True.
            _check_return_type (bool): specifies if type checking
                should be done one the data received from the server.
                Default is True.
            _content_type (str/None): force body content-type.
                Default is None and content-type will be predicted by allowed
                content-types and body.
            _host_index (int/None): specifies the index of the server
                that we want to use.
                Default is read from the configuration.
            async_req (bool): execute request asynchronously

        Returns:
            [Project]
                If the method is called asynchronously, returns the request
                thread.
        """
        kwargs['async_req'] = kwargs.get(
            'async_req', False
        )
        kwargs['_return_http_data_only'] = kwargs.get(
            '_return_http_data_only', True
        )
        kwargs['_preload_content'] = kwargs.get(
            '_preload_content', True
        )
        kwargs['_request_timeout'] = kwargs.get(
            '_request_timeout', None
        )
        kwargs['_check_input_type'] = kwargs.get(
            '_check_input_type', True
        )
        kwargs['_check_return_type'] = kwargs.get(
            '_check_return_type', True
        )
        kwargs['_content_type'] = kwargs.get(
            '_content_type')
        kwargs['_host_index'] = kwargs.get('_host_index')
        return self.get_all_projects_endpoint.call_with_http_info(**kwargs)

    def get_my_projects(
        self,
        **kwargs
    ):
        """Get My Projects  # noqa: E501

        Get projects I have access to  # noqa: E501
        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please pass async_req=True

        >>> thread = api.get_my_projects(async_req=True)
        >>> result = thread.get()


        Keyword Args:
            _return_http_data_only (bool): response data without head status
                code and headers. Default is True.
            _preload_content (bool): if False, the urllib3.HTTPResponse object
                will be returned without reading/decoding response data.
                Default is True.
            _request_timeout (int/float/tuple): timeout setting for this request. If
                one number provided, it will be total request timeout. It can also
                be a pair (tuple) of (connection, read) timeouts.
                Default is None.
            _check_input_type (bool): specifies if type checking
                should be done one the data sent to the server.
                Default is True.
            _check_return_type (bool): specifies if type checking
                should be done one the data received from the server.
                Default is True.
            _content_type (str/None): force body content-type.
                Default is None and content-type will be predicted by allowed
                content-types and body.
            _host_index (int/None): specifies the index of the server
                that we want to use.
                Default is read from the configuration.
            async_req (bool): execute request asynchronously

        Returns:
            [str]
                If the method is called asynchronously, returns the request
                thread.
        """
        kwargs['async_req'] = kwargs.get(
            'async_req', False
        )
        kwargs['_return_http_data_only'] = kwargs.get(
            '_return_http_data_only', True
        )
        kwargs['_preload_content'] = kwargs.get(
            '_preload_content', True
        )
        kwargs['_request_timeout'] = kwargs.get(
            '_request_timeout', None
        )
        kwargs['_check_input_type'] = kwargs.get(
            '_check_input_type', True
        )
        kwargs['_check_return_type'] = kwargs.get(
            '_check_return_type', True
        )
        kwargs['_content_type'] = kwargs.get(
            '_content_type')
        kwargs['_host_index'] = kwargs.get('_host_index')
        return self.get_my_projects_endpoint.call_with_http_info(**kwargs)

    def get_seqr_projects(
        self,
        **kwargs
    ):
        """Get Seqr Projects  # noqa: E501

        Get SM projects that should sync to seqr  # noqa: E501
        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please pass async_req=True

        >>> thread = api.get_seqr_projects(async_req=True)
        >>> result = thread.get()


        Keyword Args:
            _return_http_data_only (bool): response data without head status
                code and headers. Default is True.
            _preload_content (bool): if False, the urllib3.HTTPResponse object
                will be returned without reading/decoding response data.
                Default is True.
            _request_timeout (int/float/tuple): timeout setting for this request. If
                one number provided, it will be total request timeout. It can also
                be a pair (tuple) of (connection, read) timeouts.
                Default is None.
            _check_input_type (bool): specifies if type checking
                should be done one the data sent to the server.
                Default is True.
            _check_return_type (bool): specifies if type checking
                should be done one the data received from the server.
                Default is True.
            _content_type (str/None): force body content-type.
                Default is None and content-type will be predicted by allowed
                content-types and body.
            _host_index (int/None): specifies the index of the server
                that we want to use.
                Default is read from the configuration.
            async_req (bool): execute request asynchronously

        Returns:
            bool, date, datetime, dict, float, int, list, str, none_type
                If the method is called asynchronously, returns the request
                thread.
        """
        kwargs['async_req'] = kwargs.get(
            'async_req', False
        )
        kwargs['_return_http_data_only'] = kwargs.get(
            '_return_http_data_only', True
        )
        kwargs['_preload_content'] = kwargs.get(
            '_preload_content', True
        )
        kwargs['_request_timeout'] = kwargs.get(
            '_request_timeout', None
        )
        kwargs['_check_input_type'] = kwargs.get(
            '_check_input_type', True
        )
        kwargs['_check_return_type'] = kwargs.get(
            '_check_return_type', True
        )
        kwargs['_content_type'] = kwargs.get(
            '_content_type')
        kwargs['_host_index'] = kwargs.get('_host_index')
        return self.get_seqr_projects_endpoint.call_with_http_info(**kwargs)

    def update_project(
        self,
        project,
        body,
        **kwargs
    ):
        """Update Project  # noqa: E501

        Update a project by project name  # noqa: E501
        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please pass async_req=True

        >>> thread = api.update_project(project, body, async_req=True)
        >>> result = thread.get()

        Args:
            project (str):
            body ({str: (bool, date, datetime, dict, float, int, list, str, none_type)}):

        Keyword Args:
            _return_http_data_only (bool): response data without head status
                code and headers. Default is True.
            _preload_content (bool): if False, the urllib3.HTTPResponse object
                will be returned without reading/decoding response data.
                Default is True.
            _request_timeout (int/float/tuple): timeout setting for this request. If
                one number provided, it will be total request timeout. It can also
                be a pair (tuple) of (connection, read) timeouts.
                Default is None.
            _check_input_type (bool): specifies if type checking
                should be done one the data sent to the server.
                Default is True.
            _check_return_type (bool): specifies if type checking
                should be done one the data received from the server.
                Default is True.
            _content_type (str/None): force body content-type.
                Default is None and content-type will be predicted by allowed
                content-types and body.
            _host_index (int/None): specifies the index of the server
                that we want to use.
                Default is read from the configuration.
            async_req (bool): execute request asynchronously

        Returns:
            bool, date, datetime, dict, float, int, list, str, none_type
                If the method is called asynchronously, returns the request
                thread.
        """
        kwargs['async_req'] = kwargs.get(
            'async_req', False
        )
        kwargs['_return_http_data_only'] = kwargs.get(
            '_return_http_data_only', True
        )
        kwargs['_preload_content'] = kwargs.get(
            '_preload_content', True
        )
        kwargs['_request_timeout'] = kwargs.get(
            '_request_timeout', None
        )
        kwargs['_check_input_type'] = kwargs.get(
            '_check_input_type', True
        )
        kwargs['_check_return_type'] = kwargs.get(
            '_check_return_type', True
        )
        kwargs['_content_type'] = kwargs.get(
            '_content_type')
        kwargs['_host_index'] = kwargs.get('_host_index')
        kwargs['project'] = \
            project
        kwargs['body'] = \
            body
        return self.update_project_endpoint.call_with_http_info(**kwargs)

    def update_project_members(
        self,
        project,
        project_member_update,
        **kwargs
    ):
        """Update Project Members  # noqa: E501

        Update project members for specific read / write group. Not that this is protected by access to a specific access group  # noqa: E501
        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please pass async_req=True

        >>> thread = api.update_project_members(project, project_member_update, async_req=True)
        >>> result = thread.get()

        Args:
            project (str):
            project_member_update ([ProjectMemberUpdate]):

        Keyword Args:
            _return_http_data_only (bool): response data without head status
                code and headers. Default is True.
            _preload_content (bool): if False, the urllib3.HTTPResponse object
                will be returned without reading/decoding response data.
                Default is True.
            _request_timeout (int/float/tuple): timeout setting for this request. If
                one number provided, it will be total request timeout. It can also
                be a pair (tuple) of (connection, read) timeouts.
                Default is None.
            _check_input_type (bool): specifies if type checking
                should be done one the data sent to the server.
                Default is True.
            _check_return_type (bool): specifies if type checking
                should be done one the data received from the server.
                Default is True.
            _content_type (str/None): force body content-type.
                Default is None and content-type will be predicted by allowed
                content-types and body.
            _host_index (int/None): specifies the index of the server
                that we want to use.
                Default is read from the configuration.
            async_req (bool): execute request asynchronously

        Returns:
            bool, date, datetime, dict, float, int, list, str, none_type
                If the method is called asynchronously, returns the request
                thread.
        """
        kwargs['async_req'] = kwargs.get(
            'async_req', False
        )
        kwargs['_return_http_data_only'] = kwargs.get(
            '_return_http_data_only', True
        )
        kwargs['_preload_content'] = kwargs.get(
            '_preload_content', True
        )
        kwargs['_request_timeout'] = kwargs.get(
            '_request_timeout', None
        )
        kwargs['_check_input_type'] = kwargs.get(
            '_check_input_type', True
        )
        kwargs['_check_return_type'] = kwargs.get(
            '_check_return_type', True
        )
        kwargs['_content_type'] = kwargs.get(
            '_content_type')
        kwargs['_host_index'] = kwargs.get('_host_index')
        kwargs['project'] = \
            project
        kwargs['project_member_update'] = \
            project_member_update
        return self.update_project_members_endpoint.call_with_http_info(**kwargs)




#!/usr/bin/env python3
import logging

import click

from metamist.parser.generic_metadata_parser import GenericMetadataParser, run_as_sync
from metamist.parser.generic_parser import DefaultSequencing, SingleRow

PARTICIPANT_COL_NAME = 'individual_id'
SAMPLE_ID_COL_NAME = 'sample_id'
READS_COL_NAME = 'filenames'
SEQ_TYPE_COL_NAME = 'type'
CHECKSUM_COL_NAME = 'checksum'
SEQ_FACILITY_COL_NAME = 'sequencing_facility'
SEQ_LIBRARY_COL_NAME = 'sequencing_library'
READ_END_TYPE_COL_NAME = 'read_end_type'
READ_LENGTH_COL_NAME = 'read_length'

KeyMap = {
    PARTICIPANT_COL_NAME: [
        'individual id',
        'individual',
        'individual_id',
        'participant',
        'participant-id',
        'participant_id',
        'participant id',
        'agha_study_id',
    ],
    SAMPLE_ID_COL_NAME: ['sample_id', 'sample', 'sample id'],
    READS_COL_NAME: ['filename', 'filenames', 'files', 'file'],
    SEQ_TYPE_COL_NAME: ['type', 'types', 'sequencing type', 'sequencing_type'],
    SEQ_FACILITY_COL_NAME: ['facility', 'sequencing facility', 'sequencing_facility'],
    SEQ_LIBRARY_COL_NAME: [
        'library',
        'library_prep',
        'library prep',
        'library type',
        'library_type',
        'sequencing_library',
        'sequencing library',
    ],
    READ_END_TYPE_COL_NAME: [
        'read_end_type',
        'read end type',
        'read_end_types',
        'read end types',
        'end type',
        'end_type',
        'end_types',
        'end types',
    ],
    READ_LENGTH_COL_NAME: [
        'length',
        'read length',
        'read_length',
        'read lengths',
        'read_lengths',
    ],
    CHECKSUM_COL_NAME: ['md5', 'checksum'],
}

required_keys = [SAMPLE_ID_COL_NAME, READS_COL_NAME]

__DOC = """
The SampleFileMapParser is used for parsing files with format:

- ['Individual ID']
- 'Sample ID'
- 'Filenames'
- ['Type']
- 'Checksum'
- ['Sequencing Facility'] - needed for exome & rna samples
- ['Library Type'] - needed for exome & rna samples
- ['End Type'] - needed for rna samples
- ['Read Length'] - needed for rna samples

e.g.
    Sample ID       Filenames
    <sample-id>     <sample-id>.filename-R1.fastq.gz,<sample-id>.filename-R2.fastq.gz
    # OR
    <sample-id2>    <sample-id2>.filename-R1.fastq.gz
    <sample-id2>    <sample-id2>.filename-R2.fastq.gz

Example with optional columns
Note: Individual ID column must contain values in every row
Note: Any missing values in Type will default to the default_sequencing_type ('genome')
e.g.
    Individual ID	Sample ID	    Filenames	                                                                    Type
    Demeter	        sample_id001	sample_id001.filename-R1.fastq.gz,sample_id001.filename-R2.fastq.gz	            WGS
    Demeter	        sample_id001	sample_id001.exome.filename-R1.fastq.gz,sample_id001.exome.filename-R2.fastq.gz	WES
    Apollo	        sample_id002	sample_id002.filename-R1.fastq.gz	                                            WGS
    Apollo	        sample_id002	sample_id002.filename-R2.fastq.gz	                                            WGS
    Athena	        sample_id003	sample_id003.filename-R1.fastq.gz
    Athena	        sample_id003	sample_id003.filename-R2.fastq.gz
    Apollo	        sample_id004	sample_id004.filename-R1.fastq.gz
    Apollo	        sample_id004	sample_id004.filename-R2.fastq.gz

Example with optional columns for RNA samples
e.g.
    Individual ID	Sample ID	    Filenames	                                                            Type        Facility  Library    End Type  Read Length
    Hera            sample_id001	sample_id001_TSStrtRNA_R1.fastq.gz,sample_id001_TSStrtRNA_R2.fastq.gz	totalrna    VCGS      TSStrtRNA  paired    151
    Hestia          sample_id002	sample_id002_TSStrmRNA_R1.fastq.gz,sample_id002_TSStrmRNA_R2.fastq.gz	polyarna    VCGS      TSStrmRNA  paired    151


This format is useful for ingesting filenames for the seqr loading pipeline
"""

logger = logging.getLogger(__file__)
logger.addHandler(logging.StreamHandler())
logger.setLevel(logging.INFO)


class SampleFileMapParser(GenericMetadataParser):
    """Parser for SampleFileMap"""

    def __init__(
        self,
        search_locations: list[str],
        project: str,
        default_sample_type='blood',
        default_sequencing=DefaultSequencing(
            seq_type='genome', technology='short-read', platform='illumina'
        ),
        default_read_end_type: str | None = None,
        default_read_length: str | int | None = None,
        allow_extra_files_in_search_path=False,
        default_reference_assembly_location: str | None = None,
        verbose=True,
        **kwargs,
    ):
        super().__init__(
            search_locations=search_locations,
            project=project,
            participant_primary_eid_column=PARTICIPANT_COL_NAME,
            sample_primary_eid_column=SAMPLE_ID_COL_NAME,
            reads_column=READS_COL_NAME,
            checksum_column=CHECKSUM_COL_NAME,
            seq_type_column=SEQ_TYPE_COL_NAME,
            seq_facility_column=SEQ_FACILITY_COL_NAME,
            seq_library_column=SEQ_LIBRARY_COL_NAME,
            read_end_type_column=READ_END_TYPE_COL_NAME,
            read_length_column=READ_LENGTH_COL_NAME,
            default_sample_type=default_sample_type,
            default_sequencing=default_sequencing,
            default_read_end_type=default_read_end_type,
            default_read_length=default_read_length,
            default_reference_assembly_location=default_reference_assembly_location,
            allow_extra_files_in_search_path=allow_extra_files_in_search_path,
            key_map=KeyMap,
            verbose=verbose,
            **kwargs,
        )

    def get_primary_sample_id(self, row: SingleRow) -> str:
        """Get external sample ID from row"""

        if self.sample_primary_eid_column and self.sample_primary_eid_column in row:
            return row[self.sample_primary_eid_column]

        return self.get_primary_participant_id(row)

    @staticmethod
    def get_info() -> tuple[str, str]:
        """
        Information about parser, including short name and version
        """
        return ('sfmp', 'v1')


@click.command(help=__DOC)
@click.option(
    '--project',
    help='The metamist project to import manifest into',
)
@click.option('--default-sample-type', default='blood')
@click.option('--default-sequencing-type', default='wgs')
@click.option('--default-sequencing-technology', default='short-read')
@click.option('--default-sequencing-facility', default=None)
@click.option('--default-sequencing-library', default=None)
@click.option('--default-read-end-type', default=None)
@click.option('--default-read-length', default=None)
@click.option(
    '--default-reference-assembly',
    required=False,
    help=(
        'CRAMs require a reference assembly to realign. '
        'This must be provided if any of the reads are crams'
    ),
)
@click.option(
    '--search-path',
    multiple=True,
    required=True,
    help='Search path to search for files within',
)
@click.option(
    '--allow-extra-files-in-search-path',
    is_flag=True,
    help='By default, this parser will fail if there are crams, bams, fastqs '
    'in the search path that are not covered by the sample map.',
)
@click.option(
    '--confirm', is_flag=True, help='Confirm with user input before updating server'
)
@click.option(
    '--dry-run', is_flag=True, help='Just prepare the run, without comitting it'
)
@click.option('--verbose', '-v', is_flag=True, help='Verbose output')
@click.argument('manifests', nargs=-1)
@run_as_sync
async def main(  # pylint: disable=too-many-arguments
    manifests,
    search_path: list[str],
    project,
    default_sample_type='blood',
    default_sequencing_type='genome',
    default_sequencing_technology='short-read',
    default_sequencing_platform='illumina',
    default_sequencing_facility: str | None = None,
    default_sequencing_library: str | None = None,
    default_read_end_type: str | None = None,
    default_read_length: str | None = None,
    default_reference_assembly: str | None = None,
    allow_extra_files_in_search_path=False,
    confirm=False,
    dry_run=False,
    verbose=False,
):
    """Run script from CLI arguments"""
    if not manifests:
        raise ValueError('Expected at least 1 manifest')

    extra_search_paths = [m for m in manifests if m.startswith('gs://')]
    if extra_search_paths:
        search_path = list(set(search_path).union(set(extra_search_paths)))

    parser = SampleFileMapParser(
        project=project,
        default_sample_type=default_sample_type,
        default_sequencing=DefaultSequencing(
            seq_type=default_sequencing_type,
            technology=default_sequencing_technology,
            platform=default_sequencing_platform,
            facility=default_sequencing_facility,
            library=default_sequencing_library,
        ),
        default_read_end_type=default_read_end_type,
        default_read_length=default_read_length,
        default_reference_assembly_location=default_reference_assembly,
        search_locations=search_path,
        allow_extra_files_in_search_path=allow_extra_files_in_search_path,
        verbose=verbose,
    )
    for manifest in manifests:
        logger.info(f'Importing {manifest}')
        resp = await parser.from_manifest_path(
            manifest=manifest,
            confirm=confirm,
            dry_run=dry_run,
        )
        print(resp)


if __name__ == '__main__':
    # pylint: disable=no-value-for-parameter
    main()



from io import StringIO
from test.testbase import DbIsolatedTest, run_as_sync
from unittest.mock import patch

from metamist.parser.generic_parser import DefaultSequencing, ParsedParticipant
from metamist.parser.sample_file_map_parser import SampleFileMapParser


class TestSampleMapParser(DbIsolatedTest):
    """Test the TestSampleMapParser"""

    @run_as_sync
    @patch('metamist.parser.generic_parser.query_async')
    async def test_single_row_fastq(self, mock_graphql_query):
        """
        Test importing a single row, forms objects and checks response
        - MOCKS: query_async
        """
        mock_graphql_query.side_effect = self.run_graphql_query_async

        rows = [
            'Individual ID\tFilenames',
            '<sample-id>\t<sample-id>.filename-R1.fastq.gz,<sample-id>.filename-R2.fastq.gz',
        ]
        parser = SampleFileMapParser(
            search_locations=[],
            project=self.project_name,
            default_sequencing=DefaultSequencing(),
        )
        fs = ['<sample-id>.filename-R1.fastq.gz', '<sample-id>.filename-R2.fastq.gz']
        parser.filename_map = {k: 'gs://BUCKET/FAKE/' + k for k in fs}
        parser.skip_checking_gcs_objects = True

        file_contents = '\n'.join(rows)
        summary, participants = await parser.parse_manifest(
            StringIO(file_contents), delimiter='\t', dry_run=True
        )

        self.assertEqual(1, summary.participants.insert)
        self.assertEqual(0, summary.participants.update)
        self.assertEqual(1, summary.samples.insert)
        self.assertEqual(0, summary.samples.update)
        self.assertEqual(1, summary.sequencing_groups.insert)
        self.assertEqual(0, summary.sequencing_groups.update)
        self.assertEqual(1, summary.assays.insert)
        self.assertEqual(0, summary.assays.update)

        assay = participants[0].samples[0].sequencing_groups[0].assays[0]

        self.assertDictEqual({}, participants[0].samples[0].meta)
        expected_sequence_dict = {
            'reads': [
                {
                    'location': 'gs://BUCKET/FAKE/<sample-id>.filename-R1.fastq.gz',
                    'basename': '<sample-id>.filename-R1.fastq.gz',
                    'class': 'File',
                    'checksum': None,
                    'size': None,
                    'datetime_added': None,
                },
                {
                    'location': 'gs://BUCKET/FAKE/<sample-id>.filename-R2.fastq.gz',
                    'basename': '<sample-id>.filename-R2.fastq.gz',
                    'class': 'File',
                    'checksum': None,
                    'size': None,
                    'datetime_added': None,
                },
            ],
            'reads_type': 'fastq',
            'sequencing_type': 'genome',
            'sequencing_technology': 'short-read',
            'sequencing_platform': 'illumina',
        }
        self.maxDiff = None
        self.assertDictEqual(expected_sequence_dict, assay.meta)

    @run_as_sync
    @patch('metamist.parser.generic_parser.query_async')
    async def test_to_external(self, mock_graphql_query):
        """
        Test importing a single row, forms objects and checks response
        - MOCKS: query_async
        """
        mock_graphql_query.side_effect = self.run_graphql_query_async

        rows = [
            'Individual ID\tFilenames',
            '<sample-id>\t<sample-id>.filename-R1.fastq.gz,<sample-id>.filename-R2.fastq.gz',
        ]
        parser = SampleFileMapParser(
            search_locations=[],
            project=self.project_name,
            default_sequencing=DefaultSequencing(),
        )
        fs = ['<sample-id>.filename-R1.fastq.gz', '<sample-id>.filename-R2.fastq.gz']
        parser.filename_map = {k: 'gs://BUCKET/FAKE/' + k for k in fs}
        parser.skip_checking_gcs_objects = True

        file_contents = '\n'.join(rows)
        participants: list[ParsedParticipant]
        _, participants = await parser.parse_manifest(
            StringIO(file_contents), delimiter='\t', dry_run=True
        )
        for p in participants:
            p.to_sm()

    @run_as_sync
    @patch('metamist.parser.generic_parser.query_async')
    async def test_two_rows_with_provided_checksums(self, mock_graphql_query):
        """
        Test importing a single row, forms objects and checks response
        - MOCKS: get_sample_id_map_by_external, get_sequence_ids_for_sample_ids_by_type
        """
        mock_graphql_query.side_effect = self.run_graphql_query_async

        rows = [
            'Individual ID\tFilenames\tChecksum',
            '<sample-id>\t<sample-id>.filename-R1.fastq.gz,<sample-id>.filename-R2.fastq.gz\t<checksum>,<checksum2>',
            '<sample-id2>\t<sample-id2>.filename-R1.fastq.gz\t<checksum3>',
            '<sample-id2>\t<sample-id2>.filename-R2.fastq.gz\t<checksum4>',
        ]
        parser = SampleFileMapParser(
            search_locations=[],
            # doesn't matter, we're going to mock the call anyway
            project=self.project_name,
        )
        fs = [
            '<sample-id>.filename-R1.fastq.gz',
            '<sample-id>.filename-R2.fastq.gz',
            '<sample-id2>.filename-R1.fastq.gz',
            '<sample-id2>.filename-R2.fastq.gz',
        ]
        parser.filename_map = {k: 'gs://BUCKET/FAKE/' + k for k in fs}
        parser.skip_checking_gcs_objects = True

        file_contents = '\n'.join(rows)
        summary, participants = await parser.parse_manifest(
            StringIO(file_contents), delimiter='\t', dry_run=True
        )

        self.assertEqual(2, summary.participants.insert)
        self.assertEqual(0, summary.participants.update)
        self.assertEqual(2, summary.samples.insert)
        self.assertEqual(0, summary.samples.update)
        self.assertEqual(2, summary.assays.insert)
        self.assertEqual(0, summary.assays.update)
        self.maxDiff = None

        self.assertDictEqual({}, participants[0].samples[0].meta)
        expected_assay1_reads = [
            {
                'location': 'gs://BUCKET/FAKE/<sample-id>.filename-R1.fastq.gz',
                'basename': '<sample-id>.filename-R1.fastq.gz',
                'class': 'File',
                'checksum': '<checksum>',
                'size': None,
                'datetime_added': None,
            },
            {
                'location': 'gs://BUCKET/FAKE/<sample-id>.filename-R2.fastq.gz',
                'basename': '<sample-id>.filename-R2.fastq.gz',
                'class': 'File',
                'checksum': '<checksum2>',
                'size': None,
                'datetime_added': None,
            },
        ]

        self.assertListEqual(
            expected_assay1_reads,
            participants[0].samples[0].sequencing_groups[0].assays[0].meta['reads'],
        )

        expected_assay2_reads = [
            {
                'location': 'gs://BUCKET/FAKE/<sample-id2>.filename-R1.fastq.gz',
                'basename': '<sample-id2>.filename-R1.fastq.gz',
                'class': 'File',
                'checksum': '<checksum3>',
                'size': None,
                'datetime_added': None,
            },
            {
                'location': 'gs://BUCKET/FAKE/<sample-id2>.filename-R2.fastq.gz',
                'basename': '<sample-id2>.filename-R2.fastq.gz',
                'class': 'File',
                'checksum': '<checksum4>',
                'size': None,
                'datetime_added': None,
            },
        ]
        self.assertListEqual(
            expected_assay2_reads,
            participants[1].samples[0].sequencing_groups[0].assays[0].meta['reads'],
        )

    @run_as_sync
    @patch('metamist.parser.generic_parser.query_async')
    async def test_valid_rna_rows(self, mock_graphql_query):
        """
        Test importing a single row of rna data
        """

        mock_graphql_query.side_effect = self.run_graphql_query_async

        rows = [
            'Sample ID\tFilenames\tType\tfacility\tlibrary\tend_type\tread_length',
            '<sample-id>\t<sample-id>.filename-R1.fastq.gz,<sample-id>.filename-R2.fastq.gz\tpolyarna\tVCGS\tTSStrmRNA\tpaired\t151',
            '<sample-id2>\t<sample-id2>.filename-R1.fastq.gz\ttotalrna\tVCGS\tTSStrtRNA\tpaired\t151',
            '<sample-id2>\t<sample-id2>.filename-R2.fastq.gz\ttotalrna\tVCGS\tTSStrtRNA\tpaired\t151',
        ]

        parser = SampleFileMapParser(
            search_locations=[],
            # doesn't matter, we're going to mock the call anyway
            project=self.project_name,
        )
        fs = [
            '<sample-id>.filename-R1.fastq.gz',
            '<sample-id>.filename-R2.fastq.gz',
            '<sample-id2>.filename-R1.fastq.gz',
            '<sample-id2>.filename-R2.fastq.gz',
        ]
        parser.filename_map = {k: 'gs://BUCKET/FAKE/' + k for k in fs}
        parser.skip_checking_gcs_objects = True

        file_contents = '\n'.join(rows)
        summary, samples = await parser.parse_manifest(
            StringIO(file_contents), delimiter='\t', dry_run=True
        )

        self.assertEqual(0, summary.participants.insert)
        self.assertEqual(0, summary.participants.update)
        self.assertEqual(2, summary.samples.insert)
        self.assertEqual(0, summary.samples.update)
        self.assertEqual(2, summary.assays.insert)
        self.assertEqual(0, summary.assays.update)
        self.maxDiff = None

        self.assertEqual('polyarna', samples[0].sequencing_groups[0].sequencing_type)
        expected_sg1_meta = {
            'sequencing_facility': 'VCGS',
            'sequencing_library': 'TSStrmRNA',
            'read_end_type': 'paired',
            'read_length': 151,
        }
        self.assertDictEqual(
            expected_sg1_meta,
            samples[0].sequencing_groups[0].meta,
        )

        self.assertEqual('totalrna', samples[1].sequencing_groups[0].sequencing_type)
        expected_sg2_meta = {
            'sequencing_facility': 'VCGS',
            'sequencing_library': 'TSStrtRNA',
            'read_end_type': 'paired',
            'read_length': 151,
        }
        self.assertDictEqual(
            expected_sg2_meta,
            samples[1].sequencing_groups[0].meta,
        )

    @run_as_sync
    @patch('metamist.parser.generic_parser.query_async')
    async def test_invalid_rna_row(self, mock_graphql_query):
        """
        Test importing a single row of rna data
        """

        mock_graphql_query.side_effect = self.run_graphql_query_async

        rows = [
            'Sample ID\tFilenames\tType',
            '<sample-id>\t<sample-id>.filename-R1.fastq.gz,<sample-id>.filename-R2.fastq.gz\tpolyarna',
        ]

        parser = SampleFileMapParser(
            search_locations=[],
            # doesn't matter, we're going to mock the call anyway
            project=self.project_name,
        )
        fs = [
            '<sample-id>.filename-R1.fastq.gz',
            '<sample-id>.filename-R2.fastq.gz',
        ]
        parser.filename_map = {k: 'gs://BUCKET/FAKE/' + k for k in fs}
        parser.skip_checking_gcs_objects = True

        file_contents = '\n'.join(rows)
        with self.assertRaises(ValueError):
            _, _ = await parser.parse_manifest(
                StringIO(file_contents), delimiter='\t', dry_run=True
            )

    @run_as_sync
    @patch('metamist.parser.generic_parser.query_async')
    async def test_rna_row_with_default_field_values(self, mock_graphql_query):
        """
        Test importing a single row of rna data
        """

        mock_graphql_query.side_effect = self.run_graphql_query_async

        rows = [
            'Sample ID\tFilenames\tType',
            '<sample-id>\t<sample-id>.filename-R1.fastq.gz,<sample-id>.filename-R2.fastq.gz\tpolyarna',
        ]

        parser = SampleFileMapParser(
            search_locations=[],
            # doesn't matter, we're going to mock the call anyway
            project=self.project_name,
            default_sequencing=DefaultSequencing(facility='VCGS', library='TSStrmRNA'),
            default_read_end_type='paired',
            default_read_length=151,
        )
        fs = [
            '<sample-id>.filename-R1.fastq.gz',
            '<sample-id>.filename-R2.fastq.gz',
        ]
        parser.filename_map = {k: 'gs://BUCKET/FAKE/' + k for k in fs}
        parser.skip_checking_gcs_objects = True

        file_contents = '\n'.join(rows)
        summary, samples = await parser.parse_manifest(
            StringIO(file_contents), delimiter='\t', dry_run=True
        )

        self.assertEqual(0, summary.participants.insert)
        self.assertEqual(0, summary.participants.update)
        self.assertEqual(1, summary.samples.insert)
        self.assertEqual(0, summary.samples.update)
        self.assertEqual(1, summary.assays.insert)
        self.assertEqual(0, summary.assays.update)
        self.maxDiff = None

        self.assertEqual('polyarna', samples[0].sequencing_groups[0].sequencing_type)
        expected_sg1_meta = {
            'sequencing_facility': 'VCGS',
            'sequencing_library': 'TSStrmRNA',
            'read_end_type': 'paired',
            'read_length': 151,
        }
        self.assertDictEqual(
            expected_sg1_meta,
            samples[0].sequencing_groups[0].meta,
        )



# pylint: disable=protected-access
import datetime
from test.testbase import run_as_sync
from test.testbqbase import BqTest
from typing import Any
from unittest import mock

import google.cloud.bigquery as bq

from db.python.tables.bq.billing_ar_batch import BillingArBatchTable
from db.python.tables.bq.billing_filter import BillingFilter
from db.python.tables.bq.generic_bq_filter import GenericBQFilter
from db.python.utils import InternalError
from models.models import BillingColumn, BillingTotalCostQueryModel


class TestBillingArBatchTable(BqTest):
    """Test BillingArBatchTable and its methods"""

    def setUp(self):
        super().setUp()

        # setup table object
        self.table_obj = BillingArBatchTable(self.connection)

    def test_query_to_partitioned_filter(self):
        """Test query to partitioned filter conversion"""

        # given
        start_date = '2023-01-01'
        end_date = '2024-01-01'
        filters: dict[BillingColumn, str | list[Any] | dict[Any, Any]] = {
            BillingColumn.TOPIC: 'TEST_TOPIC'
        }

        # expected
        expected_filter = BillingFilter(
            day=GenericBQFilter(
                gte=datetime.datetime(2023, 1, 1, 0, 0),
                lte=datetime.datetime(2024, 1, 1, 0, 0),
            ),
            topic=GenericBQFilter(eq='TEST_TOPIC'),
        )

        query = BillingTotalCostQueryModel(
            fields=[],  # not relevant for this test, but can't be null generally
            start_date=start_date,
            end_date=end_date,
            filters=filters,
        )
        filter_ = BillingArBatchTable._query_to_partitioned_filter(query)

        # BillingFilter has __eq__ method, so we can compare them directly
        self.assertEqual(expected_filter, filter_)

    def test_error_no_connection(self):
        """Test No connection exception"""

        with self.assertRaises(InternalError) as context:
            BillingArBatchTable(None)

        self.assertTrue(
            "No connection was provided to the table 'BillingArBatchTable'"
            in str(context.exception)
        )

    def test_get_table_name(self):
        """Test get_table_name"""

        # table name is set in the class
        given_table_name = 'TEST_TABLE_NAME'

        # set table name
        self.table_obj.table_name = given_table_name

        # test get table name function
        table_name = self.table_obj.get_table_name()

        self.assertEqual(given_table_name, table_name)

    @run_as_sync
    async def test_get_batches_by_ar_guid_no_data(self):
        """Test get_batches_by_ar_guid"""

        ar_guid = '1234567890'

        # mock BigQuery result
        self.bq_result.result.return_value = []

        # test get_batches_by_ar_guid function
        (start_day, end_day, batch_ids) = await self.table_obj.get_batches_by_ar_guid(
            ar_guid
        )

        self.assertEqual(None, start_day)
        self.assertEqual(None, end_day)
        self.assertEqual([], batch_ids)

    @run_as_sync
    async def test_get_batches_by_ar_guid_one_record(self):
        """Test get_batches_by_ar_guid"""

        ar_guid = '1234567890'
        given_start_day = datetime.datetime(2023, 1, 1, 0, 0)
        given_end_day = datetime.datetime(2023, 1, 1, 2, 3)

        # mock BigQuery result
        self.bq_result.result.return_value = [
            mock.MagicMock(
                spec=bq.Row,
                start_day=given_start_day,
                end_day=given_end_day,
                batch_id='Batch1234',
            )
        ]

        # test get_batches_by_ar_guid function
        (start_day, end_day, batch_ids) = await self.table_obj.get_batches_by_ar_guid(
            ar_guid
        )

        self.assertEqual(given_start_day, start_day)
        # end day is the last day + 1
        self.assertEqual(given_end_day + datetime.timedelta(days=1), end_day)
        self.assertEqual(['Batch1234'], batch_ids)

    @run_as_sync
    async def test_get_batches_by_ar_guid_two_record(self):
        """Test get_batches_by_ar_guid"""

        ar_guid = '1234567890'
        given_start_day = datetime.datetime(2023, 1, 1, 0, 0)
        given_end_day = datetime.datetime(2023, 1, 1, 2, 3)

        # mock BigQuery result
        self.bq_result.result.return_value = [
            mock.MagicMock(
                spec=bq.Row,
                start_day=given_start_day,
                end_day=given_start_day,
                batch_id='FirstBatch',
            ),
            mock.MagicMock(
                spec=bq.Row,
                start_day=given_start_day,
                end_day=given_end_day,
                batch_id='SecondBatch',
            ),
        ]

        # test get_batches_by_ar_guid function
        (start_day, end_day, batch_ids) = await self.table_obj.get_batches_by_ar_guid(
            ar_guid
        )

        self.assertEqual(given_start_day, start_day)
        # end day is the last day + 1
        self.assertEqual(given_end_day + datetime.timedelta(days=1), end_day)
        self.assertEqual(['FirstBatch', 'SecondBatch'], batch_ids)

    @run_as_sync
    async def test_get_ar_guid_by_batch_id_no_data(self):
        """Test get_ar_guid_by_batch_id"""

        batch_id = '1234567890'

        # mock BigQuery result
        self.bq_result.result.return_value = []

        # test get_ar_guid_by_batch_id function
        _, _, ar_guid = await self.table_obj.get_ar_guid_by_batch_id(batch_id)

        self.assertEqual(None, ar_guid)

    @run_as_sync
    async def test_get_ar_guid_by_batch_id_one_rec(self):
        """Test get_ar_guid_by_batch_id"""

        batch_id = '1234567890'
        expected_ar_guid = 'AR_GUID_1234'

        # mock BigQuery result
        self.bq_result.result.return_value = [
            {
                'ar_guid': expected_ar_guid,
                'start_day': datetime.datetime(2024, 1, 1, 0, 0),
                'end_day': datetime.datetime(2024, 1, 2, 0, 0),
            }
        ]

        # test get_ar_guid_by_batch_id function
        _, _, ar_guid = await self.table_obj.get_ar_guid_by_batch_id(batch_id)

        self.assertEqual(expected_ar_guid, ar_guid)



# pylint: disable=protected-access
import datetime
from test.testbase import run_as_sync
from test.testbqbase import BqTest
from unittest import mock

import google.cloud.bigquery as bq

from db.python.layers.billing import BillingLayer
from models.enums import BillingSource
from models.models import BillingColumn, BillingTotalCostQueryModel


class TestBillingLayer(BqTest):
    """Test BillingLayer and its methods"""

    def test_table_factory(self):
        """Test table_factory"""

        layer = BillingLayer(self.connection)

        # test BillingSource types
        table_obj = layer.table_factory()
        self.assertEqual('BillingDailyTable', table_obj.__class__.__name__)

        table_obj = layer.table_factory(source=BillingSource.GCP_BILLING)
        self.assertEqual('BillingGcpDailyTable', table_obj.__class__.__name__)

        table_obj = layer.table_factory(source=BillingSource.RAW)
        self.assertEqual('BillingRawTable', table_obj.__class__.__name__)

        table_obj = layer.table_factory(source=BillingSource.AGGREGATE)
        self.assertEqual('BillingDailyTable', table_obj.__class__.__name__)

        # base columns
        table_obj = layer.table_factory(
            source=BillingSource.AGGREGATE, fields=[BillingColumn.TOPIC]
        )
        self.assertEqual('BillingDailyTable', table_obj.__class__.__name__)

        table_obj = layer.table_factory(
            source=BillingSource.AGGREGATE, filters={BillingColumn.TOPIC: 'TOPIC1'}
        )
        self.assertEqual('BillingDailyTable', table_obj.__class__.__name__)

        # columns from extended view
        table_obj = layer.table_factory(
            source=BillingSource.AGGREGATE, fields=[BillingColumn.AR_GUID]
        )
        self.assertEqual('BillingDailyExtendedTable', table_obj.__class__.__name__)

        table_obj = layer.table_factory(
            source=BillingSource.AGGREGATE, filters={BillingColumn.AR_GUID: 'AR_GUID1'}
        )
        self.assertEqual('BillingDailyExtendedTable', table_obj.__class__.__name__)

    @run_as_sync
    async def test_get_gcp_projects(self):
        """Test get_gcp_projects"""

        layer = BillingLayer(self.connection)

        # test with no muckup data, should be empty
        records = await layer.get_gcp_projects()
        self.assertEqual([], records)

        # mockup BQ results
        self.bq_result.result.return_value = [
            {'gcp_project': 'PROJECT1'},
            {'gcp_project': 'PROJECT2'},
        ]

        records = await layer.get_gcp_projects()
        self.assertEqual(['PROJECT1', 'PROJECT2'], records)

    @run_as_sync
    async def test_get_topics(self):
        """Test get_topics"""

        layer = BillingLayer(self.connection)

        # test with no muckup data, should be empty
        records = await layer.get_topics()
        self.assertEqual([], records)

        # mockup BQ results
        self.bq_result.result.return_value = [
            {'topic': 'TOPIC1'},
            {'topic': 'TOPIC2'},
        ]

        records = await layer.get_topics()
        self.assertEqual(['TOPIC1', 'TOPIC2'], records)

    @run_as_sync
    async def test_get_cost_categories(self):
        """Test get_cost_categories"""

        layer = BillingLayer(self.connection)

        # test with no muckup data, should be empty
        records = await layer.get_cost_categories()
        self.assertEqual([], records)

        # mockup BQ results
        self.bq_result.result.return_value = [
            {'cost_category': 'CAT1'},
            {'cost_category': 'CAT2'},
        ]

        records = await layer.get_cost_categories()
        self.assertEqual(['CAT1', 'CAT2'], records)

    @run_as_sync
    async def test_get_skus(self):
        """Test get_skus"""

        layer = BillingLayer(self.connection)

        # test with no muckup data, should be empty
        records = await layer.get_skus()
        self.assertEqual([], records)

        # mockup BQ results
        self.bq_result.result.return_value = [
            {'sku': 'SKU1'},
            {'sku': 'SKU2'},
        ]

        records = await layer.get_skus()
        self.assertEqual(['SKU1', 'SKU2'], records)

    @run_as_sync
    async def test_get_datasets(self):
        """Test get_datasets"""

        layer = BillingLayer(self.connection)

        # test with no muckup data, should be empty
        records = await layer.get_datasets()
        self.assertEqual([], records)

        # mockup BQ results
        self.bq_result.result.return_value = [
            {'dataset': 'DATA1'},
            {'dataset': 'DATA2'},
        ]

        records = await layer.get_datasets()
        self.assertEqual(['DATA1', 'DATA2'], records)

    @run_as_sync
    async def test_get_stages(self):
        """Test get_stages"""

        layer = BillingLayer(self.connection)

        # test with no muckup data, should be empty
        records = await layer.get_stages()
        self.assertEqual([], records)

        # mockup BQ results
        self.bq_result.result.return_value = [
            {'stage': 'STAGE1'},
            {'stage': 'STAGE2'},
        ]

        records = await layer.get_stages()
        self.assertEqual(['STAGE1', 'STAGE2'], records)

    @run_as_sync
    async def test_get_sequencing_types(self):
        """Test get_sequencing_types"""

        layer = BillingLayer(self.connection)

        # test with no muckup data, should be empty
        records = await layer.get_sequencing_types()
        self.assertEqual([], records)

        # mockup BQ results
        self.bq_result.result.return_value = [
            {'sequencing_type': 'SEQ1'},
            {'sequencing_type': 'SEQ2'},
        ]

        records = await layer.get_sequencing_types()
        self.assertEqual(['SEQ1', 'SEQ2'], records)

    @run_as_sync
    async def test_get_sequencing_groups(self):
        """Test get_sequencing_groups"""

        layer = BillingLayer(self.connection)

        # test with no muckup data, should be empty
        records = await layer.get_sequencing_groups()
        self.assertEqual([], records)

        # mockup BQ results
        self.bq_result.result.return_value = [
            {'sequencing_group': 'GRP1'},
            {'sequencing_group': 'GRP2'},
        ]

        records = await layer.get_sequencing_groups()
        self.assertEqual(['GRP1', 'GRP2'], records)

    @run_as_sync
    async def test_get_compute_categories(self):
        """Test get_compute_categories"""

        layer = BillingLayer(self.connection)

        # test with no muckup data, should be empty
        records = await layer.get_compute_categories()
        self.assertEqual([], records)

        # mockup BQ results
        self.bq_result.result.return_value = [
            {'compute_category': 'CAT1'},
            {'compute_category': 'CAT2'},
        ]

        records = await layer.get_compute_categories()
        self.assertEqual(['CAT1', 'CAT2'], records)

    @run_as_sync
    async def test_get_cromwell_sub_workflow_names(self):
        """Test get_cromwell_sub_workflow_names"""

        layer = BillingLayer(self.connection)

        # test with no muckup data, should be empty
        records = await layer.get_cromwell_sub_workflow_names()
        self.assertEqual([], records)

        # mockup BQ results
        self.bq_result.result.return_value = [
            {'cromwell_sub_workflow_name': 'CROM1'},
            {'cromwell_sub_workflow_name': 'CROM2'},
        ]

        records = await layer.get_cromwell_sub_workflow_names()
        self.assertEqual(['CROM1', 'CROM2'], records)

    @run_as_sync
    async def test_get_wdl_task_names(self):
        """Test get_wdl_task_names"""

        layer = BillingLayer(self.connection)

        # test with no muckup data, should be empty
        records = await layer.get_wdl_task_names()
        self.assertEqual([], records)

        # mockup BQ results
        self.bq_result.result.return_value = [
            {'wdl_task_name': 'WDL1'},
            {'wdl_task_name': 'WDL2'},
        ]

        records = await layer.get_wdl_task_names()
        self.assertEqual(['WDL1', 'WDL2'], records)

    @run_as_sync
    async def test_get_invoice_months(self):
        """Test get_invoice_months"""

        layer = BillingLayer(self.connection)

        # test with no muckup data, should be empty
        records = await layer.get_invoice_months()
        self.assertEqual([], records)

        # mockup BQ results
        self.bq_result.result.return_value = [
            {'invoice_month': '202301'},
            {'invoice_month': '202302'},
        ]

        records = await layer.get_invoice_months()
        self.assertEqual(['202301', '202302'], records)

    @run_as_sync
    async def test_get_namespaces(self):
        """Test get_namespaces"""

        layer = BillingLayer(self.connection)

        # test with no muckup data, should be empty
        records = await layer.get_namespaces()
        self.assertEqual([], records)

        # mockup BQ results
        self.bq_result.result.return_value = [
            {'namespace': 'NAME1'},
            {'namespace': 'NAME2'},
        ]

        records = await layer.get_namespaces()
        self.assertEqual(['NAME1', 'NAME2'], records)

    @run_as_sync
    async def test_get_total_cost(self):
        """Test get_total_cost"""

        layer = BillingLayer(self.connection)

        # test inparams exceptions:
        query = BillingTotalCostQueryModel(fields=[], start_date='', end_date='')

        with self.assertRaises(ValueError) as context:
            await layer.get_total_cost(query)

        self.assertTrue('Date and Fields are required' in str(context.exception))

        # test with no muckup data, should be empty
        query = BillingTotalCostQueryModel(
            fields=[BillingColumn.TOPIC], start_date='2024-01-01', end_date='2024-01-03'
        )
        records = await layer.get_total_cost(query)
        self.assertEqual([], records)

        # get_total_cost with mockup data is tested in test/test_bq_billing_base.py
        # BillingLayer is just wrapper for BQ tables

    @run_as_sync
    async def test_get_running_cost(self):
        """Test get_running_cost"""

        layer = BillingLayer(self.connection)

        # test inparams exceptions:
        with self.assertRaises(ValueError) as context:
            await layer.get_running_cost(
                field=BillingColumn.TOPIC, invoice_month=None, source=None
            )
        self.assertTrue('Invalid invoice month' in str(context.exception))

        with self.assertRaises(ValueError) as context:
            await layer.get_running_cost(
                field=BillingColumn.TOPIC, invoice_month='2024', source=None
            )
        self.assertTrue('Invalid invoice month' in str(context.exception))

        # test with no muckup data, should be empty
        records = await layer.get_running_cost(
            field=BillingColumn.TOPIC, invoice_month='202401', source=None
        )
        self.assertEqual([], records)

        # get_running_cost with mockup data is tested in test/test_bq_billing_base.py
        # BillingLayer is just wrapper for BQ tables

    @run_as_sync
    async def test_get_cost_by_ar_guid(self):
        """
        Test get_cost_by_ar_guid
        This test only paths in the layer,
        the logic and processing is tested in test/test_bq_billing_base.py
        """

        layer = BillingLayer(self.connection)

        # ar_guid as None, return empty results
        records = await layer.get_cost_by_ar_guid(ar_guid=None)

        # return empty record
        self.assertEqual([], records)

        # dummy ar_guid, no mockup data, return empty results
        dummy_ar_guid = '12345678'
        records = await layer.get_cost_by_ar_guid(ar_guid=dummy_ar_guid)

        # return empty record
        self.assertEqual(
            [],
            records,
        )

        # mock BigQuery first query result
        given_start_day = datetime.datetime(2023, 1, 1, 0, 0)
        given_end_day = datetime.datetime(2023, 1, 1, 2, 3)
        dummy_batch_id = '12345'

        mock_rows = mock.MagicMock(spec=bq.table.RowIterator)
        mock_rows.total_rows = 1
        mock_rows.__iter__.return_value = [
            mock.MagicMock(
                spec=bq.Row,
                batch_id=dummy_batch_id,
                start_day=given_start_day,
                end_day=given_end_day,
            ),
        ]
        self.bq_result.result.return_value = mock_rows

        records = await layer.get_cost_by_ar_guid(ar_guid=dummy_ar_guid)
        # returns empty list as those were not mocked up
        # we do not need to test cost calculation here,
        # as those are tested in test/test_bq_billing_base.py
        self.assertEqual(
            [],
            records,
        )

    @run_as_sync
    async def test_get_cost_by_batch_id(self):
        """
        Test get_cost_by_batch_id
        This test only paths in the layer,
        the logic and processing is tested in test/test_bq_billing_base.py
        """

        layer = BillingLayer(self.connection)

        # ar_guid as None, return empty results
        records = await layer.get_cost_by_batch_id(batch_id=None)

        # return empty record
        self.assertEqual([], records)

        # dummy ar_guid, no mockup data, return empty results
        dummy_batch_id = '12345'
        records = await layer.get_cost_by_batch_id(batch_id=dummy_batch_id)

        # return empty record
        self.assertEqual(
            [],
            records,
        )

        # dummy batch_id, mockup ar_guid

        # mock BigQuery result
        given_start_day = datetime.datetime(2023, 1, 1, 0, 0)
        given_end_day = datetime.datetime(2023, 1, 1, 2, 3)
        dummy_batch_id = '12345'
        dummy_ar_guid = '12345678'

        mock_rows = mock.MagicMock(spec=bq.table.RowIterator)
        mock_rows.total_rows = 1
        mock_rows.__iter__.return_value = [
            mock.MagicMock(
                spec=bq.Row,
                ar_guid=dummy_ar_guid,
                batch_id=dummy_batch_id,
                start_day=given_start_day,
                end_day=given_end_day,
            ),
        ]
        self.bq_result.result.return_value = mock_rows

        records = await layer.get_cost_by_batch_id(batch_id=dummy_batch_id)
        # returns elmpty list as those were not mocked up
        # we do not need to test cost calculation here,
        # as those are tested in test/test_bq_billing_base.py
        self.assertEqual(
            [],
            records,
        )



# noqa: B006

import unittest
import unittest.mock
from collections import namedtuple
from test.testbase import run_as_sync

from metamist.audit.generic_auditor import GenericAuditor


class TestGenericAuditor(unittest.TestCase):
    """Test the audit helper functions"""

    @run_as_sync
    @unittest.mock.patch('metamist.audit.generic_auditor.query_async')
    async def test_get_participant_data_for_dataset(self, mock_query):
        """Only participants with a non-empty samples field should be returned"""
        auditor = GenericAuditor(
            dataset='dev', sequencing_types=['genome'], file_types=('fastq',)
        )
        mock_query.return_value = {
            'project': {
                'dataset': 'test',
                'participants': [
                    {
                        'id': 1,
                        'externalId': 'P01',
                        'samples': [
                            {
                                'id': 'XPG123',
                                'sequencingGroups': [
                                    {'id': 'CPGaaa', 'assays': [{'id': 1}]}
                                ],
                            }
                        ],
                    },
                    {
                        'id': 2,
                        'externalId': 'P02',
                        'samples': [
                            {
                                'id': 'XPG456',
                                'sequencingGroups': [
                                    {'id': 'CPGbbb', 'assays': [{'id': 2}, {'id': 3}]}
                                ],
                            }
                        ],
                    },
                    {'id': 3, 'externalId': 'P01', 'samples': []},
                ],
            }
        }

        expected_participants = [
            {
                'id': 1,
                'externalId': 'P01',
                'samples': [
                    {
                        'id': 'XPG123',
                        'sequencingGroups': [{'id': 'CPGaaa', 'assays': [{'id': 1}]}],
                    }
                ],
            },
            {
                'id': 2,
                'externalId': 'P02',
                'samples': [
                    {
                        'id': 'XPG456',
                        'sequencingGroups': [
                            {'id': 'CPGbbb', 'assays': [{'id': 2}, {'id': 3}]}
                        ],
                    }
                ],
            },
        ]

        participant_data = await auditor.get_participant_data_for_dataset()

        self.assertListEqual(expected_participants, participant_data)

    def test_get_assay_map_from_participants_genome(self):
        """Only genome sequences should be mapped to the sample ID"""
        auditor = GenericAuditor(
            dataset='dev', sequencing_types=['genome'], file_types=('fastq',)
        )
        participants = [
            {
                'externalId': 'EX01',
                'id': 1,
                'samples': [
                    {
                        'id': 'XPG123',
                        'externalId': 'EX01',
                        'sequencingGroups': [
                            {
                                'id': 'CPGaaa',
                                'type': 'genome',
                                'assays': [
                                    {
                                        'id': 1,
                                        'meta': {
                                            'reads': [
                                                {
                                                    'location': 'gs://cpg-dataset-main-upload/read.fq',
                                                    'size': 11,
                                                }
                                            ]
                                        },
                                    },
                                    {
                                        'id': 2,
                                        'meta': {
                                            'reads': [
                                                {
                                                    'location': 'gs://cpg-dataset-main-upload/R1.fq',
                                                    'size': 12,
                                                },
                                                {
                                                    'location': 'gs://cpg-dataset-main-upload/R2.fq',
                                                    'size': 13,
                                                },
                                            ]
                                        },
                                    },
                                    {
                                        'id': 3,
                                        'meta': {
                                            'reads': [{'location': 'test', 'size': 0}]
                                        },
                                    },
                                ],
                            },
                            {
                                'id': 'CPGbbb',
                                'type': 'exome',
                                'assays': [
                                    {
                                        'id': 4,
                                        'meta': {
                                            'reads': [
                                                {
                                                    'location': 'gs://cpg-dataset-main-upload/exome/read.fq',
                                                    'size': 14,
                                                }
                                            ]
                                        },
                                    }
                                ],
                            },
                        ],
                    },
                ],
            }
        ]

        (
            sg_sample_id_map,
            assay_sg_id_map,
            assay_paths_sizes,
        ) = auditor.get_assay_map_from_participants(participants)

        expected_sg_sample_mapping = {'CPGaaa': 'XPG123'}
        expected_assay_sg_mapping = {1: 'CPGaaa', 2: 'CPGaaa', 3: 'CPGaaa'}
        expected_read_sizes = {
            1: [
                ('gs://cpg-dataset-main-upload/read.fq', 11),
            ],
            2: [
                ('gs://cpg-dataset-main-upload/R1.fq', 12),
                ('gs://cpg-dataset-main-upload/R2.fq', 13),
            ],
            3: [
                ('test', 0),
            ],
        }

        self.assertDictEqual(sg_sample_id_map, expected_sg_sample_mapping)
        self.assertDictEqual(assay_sg_id_map, expected_assay_sg_mapping)
        self.assertDictEqual(assay_paths_sizes, expected_read_sizes)

    def test_get_assay_map_from_participants_all(self):
        """Both genome and exome sequences should be mapped to the sample IDs"""
        auditor = GenericAuditor(
            dataset='dev', sequencing_types=['genome', 'exome'], file_types=('fastq',)
        )
        participants = [
            {
                'externalId': 'EX01',
                'id': 1,
                'samples': [
                    {
                        'id': 'XPG123',
                        'externalId': 'EX01',
                        'sequencingGroups': [
                            {
                                'id': 'CPGaaa',
                                'type': 'genome',
                                'assays': [
                                    {
                                        'id': 1,
                                        'meta': {
                                            'reads': [
                                                {
                                                    'location': 'gs://cpg-dataset-main-upload/read.fq',
                                                    'size': 11,
                                                }
                                            ]
                                        },
                                    },
                                    {
                                        'id': 2,
                                        'meta': {
                                            'reads': [
                                                {
                                                    'location': 'gs://cpg-dataset-main-upload/R1.fq',
                                                    'size': 12,
                                                },
                                                {
                                                    'location': 'gs://cpg-dataset-main-upload/R2.fq',
                                                    'size': 13,
                                                },
                                            ]
                                        },
                                    },
                                    {
                                        'id': 3,
                                        'meta': {
                                            'reads': [{'location': 'test', 'size': 0}]
                                        },
                                    },
                                ],
                            },
                            {
                                'id': 'CPGbbb',
                                'type': 'exome',
                                'assays': [
                                    {
                                        'id': 4,
                                        'meta': {
                                            'reads': [
                                                {
                                                    'location': 'gs://cpg-dataset-main-upload/exome/read.fq',
                                                    'size': 14,
                                                }
                                            ]
                                        },
                                    }
                                ],
                            },
                        ],
                    },
                ],
            }
        ]
        (
            sg_sample_id_map,
            assay_sg_id_map,
            assay_paths_sizes,
        ) = auditor.get_assay_map_from_participants(participants)

        expected_sg_sample_mapping = {'CPGaaa': 'XPG123', 'CPGbbb': 'XPG123'}
        expected_assay_sg_mapping = {1: 'CPGaaa', 2: 'CPGaaa', 3: 'CPGaaa', 4: 'CPGbbb'}
        expected_read_sizes = {
            1: [
                ('gs://cpg-dataset-main-upload/read.fq', 11),
            ],
            2: [
                ('gs://cpg-dataset-main-upload/R1.fq', 12),
                ('gs://cpg-dataset-main-upload/R2.fq', 13),
            ],
            3: [
                ('test', 0),
            ],
            4: [
                ('gs://cpg-dataset-main-upload/exome/read.fq', 14),
            ],
        }

        self.assertDictEqual(sg_sample_id_map, expected_sg_sample_mapping)
        self.assertDictEqual(assay_sg_id_map, expected_assay_sg_mapping)
        self.assertDictEqual(assay_paths_sizes, expected_read_sizes)

    def test_get_sequence_mapping_error_logging(self):
        """If the sequence reads meta field maps to a raw string, logging.error triggers"""
        auditor = GenericAuditor(
            dataset='dev', sequencing_types=['genome'], file_types=('fastq',)
        )
        participants = [
            {
                'externalId': 'EX01',
                'id': 1,
                'samples': [
                    {
                        'id': 'XPG123',
                        'externalId': 'EX01',
                        'sequencingGroups': [
                            {
                                'id': 'CPGaaa',
                                'type': 'genome',
                                'assays': [
                                    {
                                        'id': 1,
                                        'meta': {
                                            'reads': [
                                                'gs://cpg-dataset-main-upload/read.fq',
                                            ]
                                        },
                                    },
                                ],
                            },
                        ],
                    },
                ],
            },
        ]
        with self.assertLogs(level='ERROR') as log:
            _, _, _ = auditor.get_assay_map_from_participants(participants)
            self.assertEqual(len(log.output), 1)
            self.assertEqual(len(log.records), 1)
            self.assertIn(
                "ERROR:root:dev :: Got <class 'str'> read for SG CPGaaa, expected dict: gs://cpg-dataset-main-upload/read.fq",
                log.output[0],
            )

    def test_get_sequence_mapping_warning_logging(self):
        """If the sequence reads meta field is missing, logging.warning triggers"""
        auditor = GenericAuditor(
            dataset='dev', sequencing_types=['genome'], file_types=('fastq',)
        )
        participants = [
            {
                'externalId': 'EX01',
                'id': 1,
                'samples': [
                    {
                        'id': 'XPG123',
                        'externalId': 'EX01',
                        'sequencingGroups': [
                            {
                                'id': 'CPGaaa',
                                'type': 'genome',
                                'assays': [
                                    {
                                        'id': 1,
                                        'meta': {},
                                    },
                                ],
                            },
                        ],
                    },
                ],
            },
        ]
        with self.assertLogs(level='WARNING') as log:
            _, _, _ = auditor.get_assay_map_from_participants(participants)
            self.assertEqual(len(log.output), 1)
            self.assertEqual(len(log.records), 1)
            self.assertIn(
                'WARNING:root:dev :: SG CPGaaa assay 1 has no reads field',
                log.output[0],
            )

    @run_as_sync
    @unittest.mock.patch('metamist.audit.generic_auditor.query_async')
    async def test_query_genome_analyses_crams(self, mock_query):
        """Test that only the genome analysis crams for a sample map dictionary are returned"""
        auditor = GenericAuditor(
            dataset='dev', sequencing_types=['genome'], file_types=('fastq',)
        )
        mock_query.side_effect = [
            {
                'sequencingGroups': [
                    {
                        'id': 'CPGaaa',
                        'type': 'genome',
                        'analyses': [
                            {
                                'id': 1,
                                'meta': {
                                    'sequencing_type': 'genome',
                                    'sample_ids': [
                                        'CPGaaa',
                                    ],
                                },
                                'output': 'gs://cpg-dataset-main/cram/CPGaaa.cram',
                            },
                        ],
                    },
                ]
            }
        ]

        test_result = await auditor.get_analysis_cram_paths_for_dataset_sgs(
            assay_sg_id_map={1: 'CPGaaa'}
        )
        expected_result = {'CPGaaa': {1: 'gs://cpg-dataset-main/cram/CPGaaa.cram'}}

        self.assertDictEqual(test_result, expected_result)

    @run_as_sync
    @unittest.mock.patch('metamist.audit.generic_auditor.query_async')
    async def test_query_genome_and_exome_analyses_crams(self, mock_query):
        """Test that both the genome and exome analysis crams for a sample map dictionary are returned"""
        auditor = GenericAuditor(
            dataset='dev', sequencing_types=['genome', 'exome'], file_types=('fastq',)
        )
        mock_query.side_effect = [
            {
                'sequencingGroups': [
                    {
                        'id': 'CPGaaa',
                        'type': 'genome',
                        'analyses': [
                            {
                                'id': 1,
                                'meta': {
                                    'sequencing_type': 'genome',
                                    'sample_ids': [
                                        'CPGaaa',
                                    ],
                                },
                                'output': 'gs://cpg-dataset-main/cram/CPGaaa.cram',
                            },
                        ],
                    },
                    {
                        'id': 'CPGbbb',
                        'type': 'exome',
                        'analyses': [
                            {
                                'id': 2,
                                'meta': {
                                    'sequencing_type': 'exome',
                                    'sample_ids': [
                                        'CPGbbb',
                                    ],
                                },
                                'output': 'gs://cpg-dataset-main/exome/cram/CPGaaa.cram',
                            },
                        ],
                    },
                ]
            },
        ]

        test_result = await auditor.get_analysis_cram_paths_for_dataset_sgs(
            assay_sg_id_map={1: 'CPGaaa', 2: 'CPGbbb'}
        )

        expected_result = {
            'CPGaaa': {1: 'gs://cpg-dataset-main/cram/CPGaaa.cram'},
            'CPGbbb': {2: 'gs://cpg-dataset-main/exome/cram/CPGaaa.cram'},
        }

        self.assertDictEqual(test_result, expected_result)

    @run_as_sync
    @unittest.mock.patch('metamist.audit.generic_auditor.query_async')
    async def test_query_broken_analyses_crams(self, mock_query):
        """
        All analysis crams must have 'sequencing_type' meta field,
        ValueError raised if not
        """
        auditor = GenericAuditor(
            dataset='dev', sequencing_types=['genome'], file_types=('fastq',)
        )
        mock_query.return_value = {
            'sequencingGroups': [
                {
                    'id': 'CPGaaa',
                    'analyses': [
                        {
                            'id': 1,
                            'meta': {
                                'sequence_type': 'genome',
                            },
                            'sample_ids': [
                                'CPGaaa',
                            ],
                            'output': '',
                        },
                    ],
                },
            ]
        }

        with self.assertRaises(ValueError):
            await auditor.get_analysis_cram_paths_for_dataset_sgs(
                assay_sg_id_map={1: 'CPGaaa'}
            )

    @run_as_sync
    @unittest.mock.patch('metamist.audit.generic_auditor.query_async')
    async def test_query_analyses_crams_warning(self, mock_query):
        """Warn if the analysis CRAM output URI is not a CRAM file"""
        auditor = GenericAuditor(
            dataset='dev', sequencing_types=['genome'], file_types=('fastq',)
        )
        mock_query.return_value = {
            'sequencingGroups': [
                {
                    'id': 'CPGaaa',
                    'analyses': [
                        {
                            'id': 1,
                            'meta': {
                                'sequencing_type': 'genome',
                            },
                            'output': 'gs://cpg-dataset-main/cram/CPGaaa.notcram',
                        },
                    ],
                }
            ]
        }

        with self.assertLogs(level='WARNING') as log:
            _ = await auditor.get_analysis_cram_paths_for_dataset_sgs(
                assay_sg_id_map={1: 'CPGaaa'}
            )
            self.assertEqual(len(log.output), 1)
            self.assertEqual(len(log.records), 1)
            self.assertIn(
                'WARNING:root:Analysis 1 invalid output path: gs://cpg-dataset-main/cram/CPGaaa.notcram',
                log.output[0],
            )

    @run_as_sync
    @unittest.mock.patch('metamist.audit.generic_auditor.query_async')
    async def test_analyses_for_sgs_without_crams(self, mock_query):
        """Log any analyses found for samples without completed CRAMs"""
        auditor = GenericAuditor(
            dataset='dev', sequencing_types=['genome'], file_types=('fastq',)
        )
        sgs_without_crams = [  # noqa: B006
            'CPGaaa',
        ]

        mock_analysis_types_query_result = {
            'enum': {
                'analysisType': [
                    'analysis-runner',
                    'cram',
                    'custom',
                    'es-index',
                    'gvcf',
                    'joint-calling',
                    'mito-cram',
                    'mito_cram',
                    'qc',
                    'sv',
                    'web',
                ]
            }
        }

        mock_sg_analyses_query_result = {
            'sequencingGroups': [
                {
                    'id': 'CPGaaa',
                    'analyses': [
                        {
                            'id': 1,
                            'meta': {'sequencing_type': 'genome', 'sample': 'CPGaaa'},
                            'output': 'gs://cpg-dataset-main/gvcf/CPGaaa.g.vcf.gz',
                            'type': 'gvcf',
                            'timestampCompleted': '2023-05-11T16:33:00',
                        }
                    ],
                }
            ]
        }

        mock_query.side_effect = [
            mock_analysis_types_query_result,
            mock_sg_analyses_query_result,
        ]

        with self.assertLogs(level='WARNING') as log:
            # catch the warning logs from here and check below
            await auditor.analyses_for_sgs_without_crams(sgs_without_crams)

            self.assertEqual(len(log.output), 1)
            self.assertEqual(len(log.records), 1)
            self.assertIn(
                "WARNING:root:dev :: SG CPGaaa missing CRAM but has analysis {'analysis_id': 1, 'analysis_type': 'gvcf', 'analysis_output': 'gs://cpg-dataset-main/gvcf/CPGaaa.g.vcf.gz', 'timestamp_completed': '2023-05-11T16:33:00'}",
                log.output[0],
            )

    @run_as_sync
    @unittest.mock.patch(
        'metamist.audit.generic_auditor.GenericAuditor.get_gcs_bucket_subdirs_to_search'
    )
    @unittest.mock.patch(
        'metamist.audit.generic_auditor.GenericAuditor.find_files_in_gcs_buckets_subdirs'
    )
    @unittest.mock.patch(
        'metamist.audit.generic_auditor.GenericAuditor.analyses_for_sgs_without_crams'
    )
    async def test_get_complete_and_incomplete_sgs(
        self,
        mock_analyses_for_sgs_without_crams,
        mock_find_files_in_gcs_buckets_subdirs,
        mock_get_gcs_bucket_subdirs,
    ):
        """Report on samples that have completed CRAMs and those that dont"""
        assay_sg_id_map = {  # noqa: B006
            1: 'CPGaaa',
            2: 'CPGbbb',
            3: 'CPGccc',
        }
        sg_cram_paths = {  # noqa: B006
            'CPGaaa': {1: 'gs://cpg-dataset-main/cram/CPGaaa.cram'},
            'CPGbbb': {2: 'gs://cpg-dataset-main/exome/cram/CPGbbb.cram'},
        }
        auditor = GenericAuditor(
            dataset='dev', sequencing_types=['genome', 'exome'], file_types=('fastq',)
        )

        mock_get_gcs_bucket_subdirs.return_value = {
            'cpg-dataset-main': ['cram', 'exome/cram']
        }
        mock_find_files_in_gcs_buckets_subdirs.return_value = [
            'gs://cpg-dataset-main/cram/CPGaaa.cram',
            'gs://cpg-dataset-main/exome/cram/CPGbbb.cram',
        ]
        mock_analyses_for_sgs_without_crams.return_value = None

        result = await auditor.get_complete_and_incomplete_sgs(
            assay_sg_id_map=assay_sg_id_map,
            sg_cram_paths=sg_cram_paths,
        )

        expected_result = {
            'complete': {'CPGaaa': 1, 'CPGbbb': 2},
            'incomplete': ['CPGccc'],
        }

        self.assertDictEqual(result, expected_result)

    @run_as_sync
    @unittest.mock.patch('metamist.audit.generic_auditor.GenericAuditor.file_size')
    @unittest.mock.patch(
        'metamist.audit.generic_auditor.GenericAuditor.find_assay_files_in_gcs_bucket'
    )
    async def test_check_for_uningested_or_moved_assays(
        self, mock_find_sequence_files_in_gcs_bucket, mock_file_size
    ):
        """
        Test 2 ingested reads, one ingested and moved read, and one uningested read
        """
        auditor = GenericAuditor(
            dataset='dev', sequencing_types=['genome'], file_types=('fastq',)
        )
        assay_reads_sizes = {  # noqa: B006
            1: [('read1.fq', 10), ('read2.fq', 11), ('dir1/read3.fq', 12)]
        }
        completed_sgs = {'CPGaaa': [1]}
        sg_sample_id_map = {'CPGaaa': 'EXT123'}
        assay_sg_id_map = {1: 'CPGaaa'}
        sample_internal_external_id_map = {'CPGaaa': 'EXT123'}
        mock_find_sequence_files_in_gcs_bucket.return_value = [
            'read1.fq',
            'read2.fq',
            'dir2/read3.fq',
            'read4.fq',
        ]

        mock_file_size.return_value = 12

        (
            uningested_sequence_paths,
            sequences_moved_paths,
            _,
        ) = await auditor.check_for_uningested_or_moved_assays(
            bucket_name='gs://cpg-test-upload',
            assay_filepaths_filesizes=assay_reads_sizes,
            completed_sgs=completed_sgs,
            sg_sample_id_map=sg_sample_id_map,
            assay_sg_id_map=assay_sg_id_map,
            sample_internal_external_id_map=sample_internal_external_id_map,
        )

        AssayReportEntry = namedtuple(
            'AssayReportEntry',
            'sg_id assay_id assay_file_path analysis_ids filesize',
        )
        self.assertEqual(uningested_sequence_paths, {'read4.fq': []})
        self.assertEqual(
            sequences_moved_paths,
            [
                AssayReportEntry(
                    sg_id='CPGaaa',
                    assay_id=1,
                    assay_file_path='dir2/read3.fq',
                    analysis_ids=[1],
                    filesize=12,
                ),
            ],
        )

    def test_get_gcs_bucket_subdirs_to_search(self):
        """
        Takes a list of paths and extracts the bucket name and subdirectory, returning all unique pairs
        of buckets/subdirectories
        """
        paths = [
            'gs://cpg-dataset-main-upload/transfer/data.cram',
            'gs://cpg-dataset-main-upload/transfer/transfer2/data2.cram',
            'gs://cpg-dataset-main-upload/transfer/transfer2/transfer3/data3.cram',
            'gs://cpg-dataset-main-upload/transfer4/data4.cram',
            'gs://cpg-dataset-main-upload/data5.cram',
        ]
        auditor = GenericAuditor(
            dataset='dev', sequencing_types=['genome'], file_types=('fastq',)
        )
        buckets_subdirs = auditor.get_gcs_bucket_subdirs_to_search(paths)

        expected_result = {
            'cpg-dataset-main-upload': [
                'transfer/',
                'transfer/transfer2/',
                'transfer/transfer2/transfer3/',
                'transfer4/',
            ]
        }

        self.assertDictEqual(buckets_subdirs, expected_result)



from test.testbase import DbIsolatedTest, run_as_sync

from pymysql.err import IntegrityError

from db.python.layers.family import FamilyLayer
from db.python.layers.participant import ParticipantLayer
from models.models import PRIMARY_EXTERNAL_ORG, ParticipantUpsertInternal


class TestParticipantFamily(DbIsolatedTest):
    """Test moving a participant from one family to another and then back"""

    @run_as_sync
    async def setUp(self) -> None:
        super().setUp()

        fl = FamilyLayer(self.connection)

        self.fid_1 = await fl.create_family(external_id='FAM01')
        self.fid_2 = await fl.create_family(external_id='FAM02')

        pl = ParticipantLayer(self.connection)
        self.pid = (
            await pl.upsert_participant(
                ParticipantUpsertInternal(
                    external_ids={PRIMARY_EXTERNAL_ORG: 'EX01'}, reported_sex=2
                )
            )
        ).id
        self.pat_pid = (
            await pl.upsert_participant(
                ParticipantUpsertInternal(
                    external_ids={PRIMARY_EXTERNAL_ORG: 'EX01_pat'}, reported_sex=1
                )
            )
        ).id
        self.mat_pid = (
            await pl.upsert_participant(
                ParticipantUpsertInternal(
                    external_ids={PRIMARY_EXTERNAL_ORG: 'EX01_mat'}, reported_sex=2
                )
            )
        ).id

        await pl.add_participant_to_family(
            family_id=self.fid_1,
            participant_id=self.pid,
            paternal_id=self.pat_pid,
            maternal_id=self.mat_pid,
            affected=2,
        )

    @run_as_sync
    async def test_get_remove_add_family_participant_data(self):
        """
        Tests getting, removing, and adding a participants family data
        """
        pl = ParticipantLayer(self.connection)

        fp_row = await pl.get_family_participant_data(
            family_id=self.fid_1, participant_id=self.pid
        )

        expected_fp_row = {
            'family_id': self.fid_1,
            'individual_id': self.pid,
            'paternal_id': self.pat_pid,
            'maternal_id': self.mat_pid,
            'sex': 2,
            'affected': 2,
            'notes': None,
        }
        self.assertDictEqual(expected_fp_row, fp_row.to_dict())

        await pl.remove_participant_from_family(
            family_id=self.fid_1, participant_id=self.pid
        )

        await pl.add_participant_to_family(
            family_id=self.fid_2,
            participant_id=self.pid,
            paternal_id=fp_row.paternal_id,
            maternal_id=fp_row.maternal_id,
            affected=fp_row.affected,
        )

        updated_fp_row = await pl.get_family_participant_data(
            family_id=self.fid_2, participant_id=self.pid
        )

        expected_updated_fp_row = {
            'family_id': self.fid_2,
            'individual_id': self.pid,
            'paternal_id': self.pat_pid,
            'maternal_id': self.mat_pid,
            'sex': 2,
            'affected': 2,
            'notes': None,
        }
        self.assertDictEqual(expected_updated_fp_row, updated_fp_row.to_dict())

        await pl.remove_participant_from_family(
            family_id=self.fid_2, participant_id=self.pid
        )

    @run_as_sync
    async def test_update_participant_family(self):
        """Tests updating a participants family data"""
        pl = ParticipantLayer(self.connection)
        await pl.update_participant_family(
            participant_id=self.pid, old_family_id=self.fid_1, new_family_id=self.fid_2
        )

        updated_fp_row = await pl.get_family_participant_data(
            family_id=self.fid_2, participant_id=self.pid
        )

        expected_updated_fp_row = {
            'family_id': self.fid_2,
            'individual_id': self.pid,
            'paternal_id': self.pat_pid,
            'maternal_id': self.mat_pid,
            'sex': 2,
            'affected': 2,
            'notes': None,
        }
        self.assertDictEqual(expected_updated_fp_row, updated_fp_row.to_dict())

        await pl.remove_participant_from_family(
            family_id=self.fid_2, participant_id=self.pid
        )

    @run_as_sync
    async def test_update_participant_to_nonexistent_family(self):
        """Tests if error is raised and transaction rolled back for nonexistent new_family_id"""
        pl = ParticipantLayer(self.connection)

        fp_row = await pl.get_family_participant_data(
            family_id=self.fid_1, participant_id=self.pid
        )
        expected_fp_row = {
            'family_id': self.fid_1,
            'individual_id': self.pid,
            'paternal_id': self.pat_pid,
            'maternal_id': self.mat_pid,
            'sex': 2,
            'affected': 2,
            'notes': None,
        }
        self.assertDictEqual(expected_fp_row, fp_row.to_dict())

        with self.assertRaises(IntegrityError):
            await pl.update_participant_family(
                participant_id=self.pid, old_family_id=self.fid_1, new_family_id=-99
            )

        rollback_fp_row = await pl.get_family_participant_data(
            family_id=self.fid_1, participant_id=self.pid
        )

        # Update transaction should rollback, so no change expected
        self.assertDictEqual(expected_fp_row, rollback_fp_row.to_dict())



# pylint: disable=R0904,too-many-instance-attributes,too-many-locals,unused-argument,wrong-import-order,unused-argument,too-many-arguments,unused-import,too-many-lines
import asyncio
import logging
import re
import shlex
from functools import reduce
from typing import Any

import click

from metamist.parser.generic_parser import (  # noqa
    DefaultSequencing,
    GenericParser,
    GroupedRow,
    ParsedAnalysis,
    ParsedAssay,
    ParsedSample,
    ParsedSequencingGroup,
    SingleRow,
    run_as_sync,
)

__DOC = """
Parse CSV / TSV manifest of arbitrary format.
This script allows you to specify HOW you want the manifest
to be mapped onto individual data.

This script loads the WHOLE file into memory

It groups rows by the sample ID, and collapses metadata from rows.

EG:
    Sample ID       sample-collection-date  depth  qc_quality  Fastqs

    <sample-id>     2021-09-16              30x    0.997       <sample-id>.filename-R1.fastq.gz,<sample-id>.filename-R2.fastq.gz

    # OR

    <sample-id2>    2021-09-16              30x    0.997       <sample-id2>.filename-R1.fastq.gz

    <sample-id2>    2021-09-16              30x    0.997       <sample-id2>.filename-R2.fastq.gz

Given the files are in a bucket called 'gs://cpg-upload-bucket/collaborator',
and we want to achieve the following:

- Import this manifest into the "$dataset" project of SM
- Map the following to `sample.meta`:
    - "sample-collection-date" -> "collection_date"
- Map the following to `assay.meta`:
    - "depth" -> "depth"
    - "qc_quality" -> "qc.quality" (ie: {"qc": {"quality": 0.997}})
- Add a qc analysis object with the following mapped `analysis.meta`:
    - "qc_quality" -> "quality"

python parse_generic_metadata.py \
    --project $dataset \
    --sample-name-column "Sample ID" \
    --reads-column "Fastqs" \
    --sample-meta-field-map "sample-collection-date" "collection_date" \
    --assay-meta-field "depth" \
    --assay-meta-field-map "qc_quality" "qc.quality" \
    --qc-meta-field-map "qc_quality" "quality" \
    --search-path "gs://cpg-upload-bucket/collaborator" \
    <manifest-path>
"""


logger = logging.getLogger(__file__)
logger.addHandler(logging.StreamHandler())
logger.setLevel(logging.INFO)

RE_FILENAME_SPLITTER = re.compile('[,;]')


class GenericMetadataParser(GenericParser):
    """Parser for GenericMetadataParser"""

    def __init__(
        self,
        project: str,
        search_locations: list[str],
        # sample columns
        sample_primary_eid_column: str,
        sample_external_id_column_map: dict[str, str] | None = None,
        # Participant columns
        participant_primary_eid_column: str | None = None,
        reported_sex_column: str | None = None,
        reported_gender_column: str | None = None,
        karyotype_column: str | None = None,
        participant_external_id_column_map: dict[str, str] | None = None,
        # Sequencing metadata columns
        seq_type_column: str | None = None,
        seq_technology_column: str | None = None,
        seq_platform_column: str | None = None,
        seq_facility_column: str | None = None,
        seq_library_column: str | None = None,
        # Assay columns
        assay_external_id_column: str | None = None,
        reads_column: str | None = None,
        checksum_column: str | None = None,
        read_end_type_column: str | None = None,
        read_length_column: str | None = None,
        reference_assembly_location_column: str | None = None,
        # GVCF columns
        gvcf_column: str | None = None,
        # Meta field key maps
        participant_meta_map: dict[str, str] | None = None,
        sample_meta_map: dict[str, str] | None = None,
        assay_meta_map: dict[str, str] | None = None,
        qc_meta_map: dict[str, str] | None = None,
        # Default values
        default_reference_assembly_location: str | None = None,
        default_sample_type: str | None = None,
        default_sequencing=DefaultSequencing(
            seq_type='genome', technology='short-read', platform='illumina'
        ),
        default_read_end_type: str | None = None,
        default_read_length: str | int | None = None,
        batch_number: str | None = None,
        allow_extra_files_in_search_path=False,
        **kwargs,
    ):
        """
        Configuration options for a GenericMetadata parser.

        Args:
            project (str): The name of the project to insert into
            search_locations (list[str]):
                A list of locations to search for unqualified files in.

            sample_primary_eid_column (str): The name of the column containing the sample name.
            sample_external_id_column_map (str | None, optional):
                {column: eid_name} mapping for sample external_ids. This should NOT
                include the sample_primary_eid_column.
            sample_meta_map (dict[str, str] | None, optional):
                A mapping of column names to keys for sample.meta.
                {[column]: [metadata_key]}
            default_sample_type (str | None, optional):
                The default sample type to use if not provided in the manifest.

            participant_primary_eid_column (str | None, optional):
                The name of the column containing the participant name. Unsupplied means
                there are no participants represented in the file.
            reported_sex_column (str | None, optional):
                The name of the column containing the reported sex, this will get parsed
                into (0 = unknown, 1 = male, 2 = female), otherwise, ValueError
            reported_gender_column (str | None, optional):
                The name of the column containing the reported gender, no validation.
            karyotype_column (str | None, optional):
                The name of the column containing the karyotype, no validation.
            participant_external_id_column_map (str | None, optional):
                {column: eid_name} mapping for participant external_ids.
                This should NOT include the participant_primary_eid_column.
            participant_meta_map (dict[str, str] | None, optional):
                A mapping of column names to keys for participant.meta.
                {[column]: [metadata_key]}

            default_sequencing (_type_, optional):
                Default sequencing values to use when not found.Defaults to
                DefaultSequencing( seq_type='genome', technology='short-read', platform='illumina' ).
            seq_type_column (str | None, optional):
                The name of the column containing the sequencing type.
            seq_technology_column (str | None, optional):
                The name of the column containing the sequencing technology.
            seq_platform_column (str | None, optional):
                The name of the column containing the sequencing platform.
            seq_facility_column (str | None, optional):
                The name of the column containing the sequencing facility.
            seq_library_column (str | None, optional):
                The name of the column containing the sequencing library.

            assay_id_column (str | None, optional):
                The name of the column containing an external assay ID to group on.
            assay_meta_map (dict[str, str] | None, optional):
                A mapping of column names to keys for assay.meta.
                {[column]: [metadata_key]}
            reads_column (str | None, optional):
                The name of the column containing the reads.
            checksum_column (str | None, optional):
                The name of the column containing the checksum.
            read_end_type_column (str | None, optional):
                The name of the column containing the read end type.
            default_read_end_type (str | None, optional):
                The default read end type to use if not provided in the manifest.
            read_length_column (str | None, optional):
                The name of the column containing the read length.
            default_read_length (str | int | None, optional):
                The default read length to use if not provided in the manifest.
            reference_assembly_location_column (str | None, optional):
                The name of the column containing the reference assembly location.
            default_reference_assembly_location (str | None, optional):
                Path to a reference_assembly where the reads_type is a CRAM

            gvcf_column (str | None, optional):
                The name of the column containing the path to a gvcf

            qc_meta_map (dict[str, str] | None, optional):
                A mapping of column names to keys for QC analysis.meta

            batch_number (str | None, optional): a batch number to assign to the samples
            allow_extra_files_in_search_path (bool, optional):
                Whether to allow files in the search path that are not in the manifest.
                This could be useful to ensure that all files uploaded to some dir
                are covered by the manifest, and there aren't extras.
        """
        super().__init__(
            path_prefix=None,
            search_paths=search_locations,
            project=project,
            default_sample_type=default_sample_type,
            default_sequencing=default_sequencing,
            default_read_end_type=default_read_end_type,
            default_read_length=default_read_length,
            **kwargs,
        )

        if not sample_primary_eid_column:
            raise ValueError('A sample name column MUST be provided')

        self.cpg_id_column = 'Internal CPG Sequencing Group ID'
        self.sample_primary_eid_column = sample_primary_eid_column
        self.sample_external_id_column_map = sample_external_id_column_map or {}

        # Participant columns
        self.participant_primary_eid_column = participant_primary_eid_column
        self.participant_external_id_column_map = (
            participant_external_id_column_map or {}
        )
        self.reported_sex_column = reported_sex_column
        self.reported_gender_column = reported_gender_column
        self.karyotype_column = karyotype_column

        # Sequencing metadata columns
        self.seq_type_column = seq_type_column
        self.seq_technology_column = seq_technology_column
        self.seq_platform_column = seq_platform_column
        self.seq_facility_column = seq_facility_column
        self.seq_library_column = seq_library_column

        # Assay columns
        self.assay_id_column = assay_external_id_column
        self.reads_column = reads_column
        self.checksum_column = checksum_column
        self.read_end_type_column = read_end_type_column
        self.read_length_column = read_length_column
        self.reference_assembly_location_column = reference_assembly_location_column

        # Meta field key maps
        self.participant_meta_map = participant_meta_map or {}
        self.sample_meta_map = sample_meta_map or {}
        self.assay_meta_map = assay_meta_map or {}
        self.qc_meta_map = qc_meta_map or {}

        # GVCF column
        self.gvcf_column = gvcf_column

        # Default values
        self.batch_number = batch_number
        self.default_reference_assembly_location = default_reference_assembly_location
        self.allow_extra_files_in_search_path = allow_extra_files_in_search_path

    def get_primary_sample_id(self, row: SingleRow) -> str:
        """Get external sample ID from row"""
        return row[self.sample_primary_eid_column].strip()

    async def get_cpg_sample_id_from_row(self, row: SingleRow) -> str | None:
        """Get internal cpg id from a row using get_sample_id and an api call"""
        return row.get(self.cpg_id_column, None)

    def get_sample_type(self, row: GroupedRow) -> str:
        """Get sample type from row"""
        if self.default_sample_type:
            return self.default_sample_type
        return None

    def get_sequencing_types(self, row: GroupedRow) -> list[str]:
        """
        Get assay types from grouped row
        if SingleRow: return assay type
        if GroupedRow: return assay types for all rows
        """
        if isinstance(row, dict):
            return [self.get_sequencing_type(row)]
        return [
            str(r.get(self.seq_type_column, self.default_sequencing.seq_type))
            for r in row
        ]

    def get_sequencing_technology(self, row: SingleRow) -> str:
        """Get assay technology for single row"""
        value = (
            row.get(self.seq_technology_column, None)
            or self.default_sequencing.technology
        )
        value = value.lower()

        if value == 'ont':
            value = 'long-read'

        return str(value)

    def get_sequencing_platform(self, row: SingleRow) -> str:
        """Get sequencing platform for single row"""
        value = (
            row.get(self.seq_platform_column, None) or self.default_sequencing.platform
        )
        value = value.lower()

        return str(value)

    def get_sequencing_type(self, row: SingleRow) -> str:
        """Get assay type from row"""
        value = row.get(self.seq_type_column, None) or self.default_sequencing.seq_type
        value = value.lower()

        if value == 'wgs':
            value = 'genome'
        elif value == 'wes':
            value = 'exome'
        elif 'mt' in value:
            value = 'mtseq'
        elif 'polya' in value or 'mrna' in value:
            value = 'polyarna'
        elif 'total' in value:
            value = 'totalrna'
        elif 'single' in value:
            value = 'singlecellrna'

        return str(value)

    def get_sequencing_facility(self, row: SingleRow) -> str:
        """Get sequencing facility from row"""
        value = (
            row.get(self.seq_facility_column, None) or self.default_sequencing.facility
        )
        value = str(value) if value else None
        return value

    def get_sequencing_library(self, row: SingleRow) -> str:
        """Get sequencing library from row"""
        value = (
            row.get(self.seq_library_column, None) or self.default_sequencing.library
        )
        value = str(value) if value else None
        return value

    def get_read_end_type(self, row: SingleRow) -> str:
        """Get read end type from row"""
        value = row.get(self.read_end_type_column, None) or self.default_read_end_type
        value = str(value).lower() if value else None
        return value

    def get_read_length(self, row: SingleRow) -> int:
        """Get read length from row"""
        value = row.get(self.read_length_column, None) or self.default_read_length
        value = int(value) if value else None
        return value

    def get_assay_id(self, row: GroupedRow) -> dict[str, str] | None:
        """Get external assay ID from row. Needs to be implemented per parser.
        NOTE: To be re-thought after assay group changes are applied"""
        return None

    def get_primary_participant_id(self, row: SingleRow) -> str | None:
        """Get external participant ID from row"""
        if (
            not self.participant_primary_eid_column
            or self.participant_primary_eid_column not in row
        ):
            raise ValueError('Participant column does not exist')
        return row[self.participant_primary_eid_column]

    def get_reported_sex(self, row: GroupedRow) -> int | None:
        """Get reported sex from grouped row"""

        if not self.reported_sex_column:
            return None

        reported_sex = row[0].get(self.reported_sex_column, None)

        if reported_sex is None:
            return None
        if reported_sex == '':
            return None
        if reported_sex.lower() == 'female':
            return 2
        if reported_sex.lower() == 'male':
            return 1

        raise ValueError(
            f'{reported_sex} could not be identified as an input for reported_sex'
        )

    def get_reported_gender(self, row: GroupedRow) -> str | None:
        """Get reported gender from grouped row"""
        return row[0].get(self.reported_gender_column, None)

    def get_karyotype(self, row: GroupedRow) -> str | None:
        """Get karyotype from grouped row"""
        return row[0].get(self.karyotype_column, None)

    def has_participants(self, rows: list[SingleRow]) -> bool:
        """Returns True if the file has a Participants column"""
        return self.participant_primary_eid_column in rows[0]

    async def validate_rows(self, rows):
        await super().validate_rows(rows)

        errors = []
        errors.extend(await self.check_files_covered_by_rows(rows))

        if errors:
            raise ValueError(', '.join(errors))

    @staticmethod
    def flatten_irregular_list(irregular_list):
        """
        Flatten an irregular list: [1, [2, 3], 4]

        >>> GenericMetadataParser.flatten_irregular_list([1, [2, 3], [4,5]])
        [1, 2, 3, 4, 5]
        """
        return (
            [
                element
                for item in irregular_list
                for element in GenericMetadataParser.flatten_irregular_list(item)
            ]
            if isinstance(irregular_list, list)
            else [irregular_list]
        )

    async def get_all_files_from_row(self, sample_id: str, row):
        """Get all files from row, to allow subparsers to include other files"""
        fns = await self.get_read_filenames(sample_id, row)

        return self.flatten_irregular_list(fns)

    async def check_files_covered_by_rows(
        self, rows: list[dict[str, Any]]
    ) -> list[str]:
        """
        Check that the files in the search_paths are completely covered by the sample_map
        """
        filename_promises = []
        for grp in rows:
            for r in grp if isinstance(grp, list) else [grp]:
                filename_promises.append(
                    self.get_all_files_from_row(self.get_primary_sample_id(r), r)
                )

        files_from_rows: list[str] = sum(await asyncio.gather(*filename_promises), [])
        filenames_from_rows = set(f.strip() for f in files_from_rows if f and f.strip())
        relevant_extensions = ('.cram', '.fastq.gz', '.fastq', 'fq.gz', '.fq', '.bam')

        # we need to explicitly filter filenames from rows not to include absolute
        # paths, otherwise the below check will flag it as missing
        absolute_path_starts = ('/', 'gs://', 'https://')
        filenames_from_rows = set(
            f
            for f in filenames_from_rows
            if not (any(f.startswith(p) for p in absolute_path_starts))
        )

        def filename_filter(f):
            return any(f.endswith(ext) for ext in relevant_extensions)

        file_from_search_paths = set(filter(filename_filter, self.filename_map.keys()))

        files_in_search_path_not_in_map = file_from_search_paths - filenames_from_rows
        missing_files = filenames_from_rows - file_from_search_paths

        errors = []

        if missing_files:
            errors.append(
                'There are files specified in the map, but not found in '
                f'the search paths: {", ".join(missing_files)}'
            )
        if files_in_search_path_not_in_map:
            m = (
                'There are files in the search path that are NOT covered by the file map: '
                f'{", ".join(files_in_search_path_not_in_map)}'
            )
            if self.allow_extra_files_in_search_path:
                logger.warning(m)
            else:
                errors.append(m)

        return errors

    @staticmethod
    def merge_dicts(a: dict, b: dict):
        """
        Recursively merge two dictionaries:
        - collapse equal values
        - put differing values into a list (not guaranteeing order)
        """
        if b is None:
            return a
        if a is None:
            return b

        res = {}
        for key in set(a.keys()).union(b.keys()):
            a_val = a.get(key)
            b_val = b.get(key)
            if a_val is not None and b_val is not None:
                # combine values
                a_is_dict = isinstance(a_val, dict)
                b_is_dict = isinstance(b_val, dict)

                if a_is_dict and b_is_dict:
                    # merge dict
                    res[key] = GenericMetadataParser.merge_dicts(a_val, b_val)
                elif a_val == b_val:
                    res[key] = a_val
                else:
                    res[key] = [a_val, b_val]
            else:
                res[key] = a_val or b_val

        return res

    @staticmethod
    def collapse_arbitrary_meta(key_map: dict[str, str], row: GroupedRow):
        """
        This is a little bit tricky

        >>> GenericMetadataParser.collapse_arbitrary_meta({'key1': 'new_key'}, {'key1': True})
        {'new_key': True}

        >>> GenericMetadataParser.collapse_arbitrary_meta({'key1': 'new_key'}, [{'key1': True}, {'key1': True}])
        {'new_key': True}

        >>> GenericMetadataParser.collapse_arbitrary_meta({'key1': 'new_key'}, [{'key1': True}, {'key1': None}])
        {'new_key': True}

        >>> GenericMetadataParser.collapse_arbitrary_meta({'key1': 'new_key'}, [{'key1': True}])
        {'new_key': True}

        >>> GenericMetadataParser.collapse_arbitrary_meta({'key1': 'new.key'}, [{'key1': True}])
        {'new': {'key': True}}

        >>> GenericMetadataParser.collapse_arbitrary_meta({'key1': 'new_key'}, [{'key1': [1,2]}, {'key1': [2,3]}])
        {'new_key': [[1, 2], [2, 3]}

        >>> GenericMetadataParser.collapse_arbitrary_meta({'key1': 'new_key'}, [{}])
        {}

        >>> GenericMetadataParser.collapse_arbitrary_meta({'key1': 'new_key'}, [{'key1': None}])
        {}

        >>> GenericMetadataParser.collapse_arbitrary_meta({'key1': 'new.key'}, [{'key1': 1}, {'key1': 2}, {'key1': 3}])
        {'new': {'key': [1, 2, 3]}}

        # multiple keys sometimes is ordered, so check the sorted(dict.items())
        >>> import json; json.dumps(GenericMetadataParser.collapse_arbitrary_meta({'key1': 'new.key', 'key2': 'new.another'}, [{'key1': 1}, {'key1': 2}, {'key2': False}]), sort_keys=True)
        '{"new": {"another": false, "key": [1, 2]}}'
        """
        if not key_map or not row:
            return {}

        def unstring_value(value: Any) -> bool:
            """Convert strings to boolean or number if possible"""
            if not isinstance(value, str):
                return value
            if value.lower() == 'true':
                return True
            if value.lower() == 'false':
                return False
            try:
                return int(value)
            except ValueError:
                try:
                    return float(value)
                except ValueError:
                    return value

        def prepare_dict_from_keys(key_parts: list[str], val):
            """Recursive production of dictionary"""
            if len(key_parts) == 1:
                return {key_parts[0]: val}
            return {key_parts[0]: prepare_dict_from_keys(key_parts[1:], val)}

        dicts = [{}]
        for row_key, dict_key in key_map.items():
            if isinstance(row, list):
                if len(row) == 1:
                    value = unstring_value(row[0].get(row_key))
                else:
                    inner_values = [
                        unstring_value(r[row_key])
                        for r in row
                        if r.get(row_key) is not None
                    ]
                    if any(isinstance(inner, list) for inner in inner_values):
                        # lists are unhashable
                        value = inner_values
                    else:
                        value = sorted(
                            set(inner_values), key=str
                        )  # sorted for unit test consistency
                        if len(value) == 0:
                            continue
                        if len(value) == 1:
                            value = unstring_value(value[0])
            else:
                if row_key not in row:
                    continue
                value = unstring_value(row[row_key])
            dicts.append(prepare_dict_from_keys(dict_key.split('.'), value))

        return reduce(GenericMetadataParser.merge_dicts, dicts)

    @staticmethod
    def process_filename_value(string: str | list[str]) -> list[str]:
        """
        Split on multiple delimiters, ;,

        >>> GenericMetadataParser.process_filename_value('Filename1-fastq.gz;Filename2.fastq.gz')
        ['Filename1-fastq.gz', 'Filename2.fastq.gz']

        >>> GenericMetadataParser.process_filename_value('Filename1-fastq.gz, Filename2.fastq.gz')
        ['Filename1-fastq.gz', 'Filename2.fastq.gz']

        >>> GenericMetadataParser.process_filename_value('Filename1 with spaces fastq.gz')
        ['Filename1 with spaces fastq.gz']

        >>> GenericMetadataParser.process_filename_value(['filename ;filename2, filename3', ' filename4'])
        ['filename', 'filename2', 'filename3', 'filename4']
        """
        if not string:
            return []
        if isinstance(string, list):
            return sorted(
                set(
                    r
                    for f in string
                    for r in GenericMetadataParser.process_filename_value(f)
                )
            )

        filenames = [f.strip() for f in RE_FILENAME_SPLITTER.split(string)]
        filenames = [f for f in filenames if f]

        whitespace_filenames = [f for f in filenames if ' ' in f]
        if whitespace_filenames:
            logger.warning(
                'Whitespace detected in filenames: '
                + ','.join(shlex.quote(str(s)) for s in whitespace_filenames)
            )

        return filenames

    async def get_read_filenames(
        self, sample_id: str | None, row: SingleRow
    ) -> list[str]:
        """Get paths to reads from a row"""
        if not self.reads_column or self.reads_column not in row:
            return []
        # more post-processing
        return self.process_filename_value(row[self.reads_column])

    async def get_checksums_from_row(
        self, sample_id: str | None, row: SingleRow, read_filenames: list[str]
    ) -> list[str | None] | None:
        """
        Get checksums for some row, you must either return:
            - no elements, or
            - number of elements equal to read_filenames

        Each element should be a string or None.
        """
        if not self.checksum_column or self.checksum_column not in row:
            return []

        return self.process_filename_value(row[self.checksum_column])

    async def get_gvcf_filenames(self, sample_id: str, row: GroupedRow) -> list[str]:
        """Get paths to gvcfs from a row"""
        if not self.gvcf_column:
            return []

        gvcf_filenames: list[str] = []
        for r in row if isinstance(row, list) else [row]:
            if self.gvcf_column in r:
                gvcf_filenames.extend(self.process_filename_value(r[self.gvcf_column]))

        return gvcf_filenames

    async def get_sample_meta_from_group(self, rows: GroupedRow):
        """Get sample-metadata from row"""
        return self.collapse_arbitrary_meta(self.sample_meta_map, rows)

    def get_participant_external_ids(self, row: GroupedRow) -> dict[str, str]:
        eids = {}
        for r in row:
            for column, name in self.participant_external_id_column_map.items():
                new_eid = r.get(column)
                if not new_eid:
                    continue
                if name in eids and new_eid != eids[name]:
                    raise ValueError
                eids[name] = new_eid

        return eids

    def get_sample_external_ids(self, row: GroupedRow) -> dict[str, str]:
        eids = {}
        for r in row:
            for column, name in self.sample_external_id_column_map.items():
                new_eid = r.get(column)
                if not new_eid:
                    continue
                if name in eids and new_eid != eids[name]:
                    raise ValueError
                eids[name] = new_eid

        return eids

    async def get_participant_meta_from_group(self, rows: GroupedRow):
        """Get participant-metadata from rows then set it in the ParticipantMetaGroup"""
        return self.collapse_arbitrary_meta(self.participant_meta_map, rows)

    async def get_sequencing_group_meta(
        self, sequencing_group: ParsedSequencingGroup
    ) -> dict:
        """Get sequencing group metadata from variant (vcf) files"""
        meta: dict[str, Any] = {}
        gvcf_filenames: list[str] = []

        for r in sequencing_group.rows:
            if self.gvcf_column and self.gvcf_column in r:
                gvcf_filenames.extend(self.process_filename_value(r[self.gvcf_column]))

        # strip in case collaborator put "file1, file2"
        full_gvcf_filenames: list[str] = []

        if gvcf_filenames:
            full_gvcf_filenames.extend(
                self.file_path(f.strip()) for f in gvcf_filenames if f.strip()
            )

        variant_file_types: dict[str, dict[str, list]] = await self.parse_files(
            sequencing_group.sample.primary_external_id, full_gvcf_filenames, None
        )
        variants: dict[str, list] = variant_file_types.get('variants')

        if variants:
            if 'gvcf' in variants:
                meta['gvcfs'] = variants.get('gvcf')
                meta['gvcf_types'] = 'gvcf'

            if 'vcf' in variants:
                meta['vcfs'] = variants['vcf']
                meta['vcf_type'] = 'vcf'

        return meta

    async def get_read_and_ref_files_and_checksums(
        self, sample_id: str, rows: GroupedRow
    ) -> tuple[list[str], list[str], set[str]]:
        """Get read filenames and checksums from rows."""
        read_filenames: list[str] = []
        read_checksums: list[str] = []
        reference_assemblies: set[str] = set()
        for r in rows:
            _rfilenames = await self.get_read_filenames(sample_id=sample_id, row=r)
            read_filenames.extend(_rfilenames)
            if self.checksum_column and self.checksum_column in r:
                checksums = await self.get_checksums_from_row(sample_id, r, _rfilenames)
                if not checksums:
                    checksums = [None] * len(_rfilenames)
                read_checksums.extend(checksums)
            if self.reference_assembly_location_column:
                ref = r.get(self.reference_assembly_location_column)
                if ref:
                    reference_assemblies.add(ref)
        return read_filenames, read_checksums, reference_assemblies

    async def parse_cram_assays(
        self, sample: ParsedSample, reference_assemblies: set[str]
    ) -> dict[str, Any]:
        """Parse CRAM assays"""
        if len(reference_assemblies) > 1:
            # sorted for consistent testing
            str_ref_assemblies = ', '.join(sorted(reference_assemblies))
            raise ValueError(
                f'Multiple reference assemblies were defined for {sample.primary_external_id}: {str_ref_assemblies}'
            )
        if len(reference_assemblies) == 1:
            ref = next(iter(reference_assemblies))
        else:
            ref = self.default_reference_assembly_location
            if not ref:
                raise ValueError(
                    f'Reads type for {sample.primary_external_id!r} is CRAM, but a reference '
                    f'is not defined, please set the default reference assembly path'
                )

        ref_fp = self.file_path(ref)
        secondary_files = await self.create_secondary_file_objects_by_potential_pattern(
            ref_fp, ['.fai']
        )
        cram_reference = await self.create_file_object(
            ref_fp, secondary_files=secondary_files
        )
        return {'reference_assembly': cram_reference}

    async def get_assays_from_group(
        self, sequencing_group: ParsedSequencingGroup
    ) -> list[ParsedAssay]:
        """Get assays from assay group + rows"""
        sample = sequencing_group.sample
        rows = sequencing_group.rows

        collapsed_assay_meta = self.collapse_arbitrary_meta(self.assay_meta_map, rows)

        assays = []

        (
            read_filenames,
            read_checksums,
            reference_assemblies,
        ) = await self.get_read_and_ref_files_and_checksums(
            sample.primary_external_id, rows
        )

        # strip in case collaborator put "file1, file2"
        full_read_filenames: list[str] = []
        if read_filenames:
            full_read_filenames.extend(
                self.file_path(f.strip()) for f in read_filenames if f.strip()
            )
        read_file_types: dict[str, dict[str, list]] = await self.parse_files(
            sample.primary_external_id, full_read_filenames, read_checksums
        )
        reads: dict[str, list] = read_file_types.get('reads')
        if not reads:
            return []

        keys = list(reads.keys())
        if len(keys) > 1:
            # 2021-12-14 mfranklin: In future we should return multiple
            #       assay meta, and handle that in the generic parser
            raise ValueError(
                f'Multiple types of reads found ({", ".join(keys)}), currently not supported'
            )

        reads_type = keys[0]
        collapsed_assay_meta['reads_type'] = reads_type
        # collapsed_assay_meta['reads'] = reads[reads_type]

        if reads_type == 'cram':
            collapsed_assay_meta.update(
                await self.parse_cram_assays(sample, reference_assemblies)
            )

        if self.batch_number is not None:
            collapsed_assay_meta['batch'] = self.batch_number

        if sequencing_group.sequencing_type in [
            'exome',
            'polyarna',
            'totalrna',
            'singlecellrna',
        ]:
            rows = sequencing_group.rows
            # Exome / RNA should have facility and library, allow missing for exomes - for now
            if self.get_sequencing_facility(rows[0]):
                collapsed_assay_meta['sequencing_facility'] = (
                    self.get_sequencing_facility(rows[0])
                )
            if self.get_sequencing_library(rows[0]):
                collapsed_assay_meta['sequencing_library'] = (
                    self.get_sequencing_library(rows[0])
                )
            # RNA requires read end type and length as well
            if sequencing_group.sequencing_type != 'exome':
                collapsed_assay_meta['read_end_type'] = self.get_read_end_type(rows[0])
                collapsed_assay_meta['read_length'] = self.get_read_length(rows[0])
                # if any of the above fields are not set for an RNA assay, raise an error
                if not all(collapsed_assay_meta.values()):
                    raise ValueError(
                        f'Not all required fields were set for RNA sample {sample.primary_external_id}:\n'
                        f'{collapsed_assay_meta}\n'
                        f'Use the default value arguments if they are not present in the manifest.'
                    )

        for read in reads[reads_type]:
            assays.append(
                ParsedAssay(
                    group=sequencing_group,
                    # we don't determine which set of rows belong to a assay,
                    # as we grab all reads, and then determine assay
                    # grouping from there.
                    rows=sequencing_group.rows,
                    internal_assay_id=None,
                    external_assay_ids={},
                    # unfortunately hard to break them up by row in the current format
                    # assay_status=self.get_assay_status(rows),
                    assay_type='sequencing',
                    meta={
                        **collapsed_assay_meta,
                        'reads': read,
                        'sequencing_type': self.get_sequencing_type(rows[0]),
                        'sequencing_technology': sequencing_group.sequencing_technology,
                        'sequencing_platform': sequencing_group.sequencing_platform,
                    },
                )
            )
        if not sequencing_group.meta:
            sequencing_group.meta = self.get_sequencing_group_meta_from_assays(assays)

        return assays

    async def get_analyses_from_sequencing_group(
        self, sequencing_group: ParsedSequencingGroup
    ) -> list[ParsedAnalysis]:
        if not self.qc_meta_map:
            return []

        sample_id = sequencing_group.sample.primary_external_id

        return [
            ParsedAnalysis(
                sequencing_group=sequencing_group,
                status=self.get_analysis_status(sample_id, sequencing_group.rows),
                type_=self.get_analysis_type(sample_id, sequencing_group.rows),
                meta=self.collapse_arbitrary_meta(
                    self.qc_meta_map, sequencing_group.rows
                ),
                rows=sequencing_group.rows,
                output=None,
            )
        ]

    @staticmethod
    def get_info() -> tuple[str, str]:
        """
        Information about parser, including short name and version
        """
        return ('gmp', 'v1')


@click.command(help=__DOC)
@click.option(
    '--project',
    required=True,
    help='The metamist project ($DATASET) to import manifest into',
)
@click.option('--search-path', multiple=True, required=True)
@click.option('--sample-name-column', required=True)
@click.option(
    '--participant-column',
    help='Column where the external participant id is held',
)
@click.option(
    '--reported-sex-column',
    help='Column where the reported sex is held',
)
@click.option(
    '--reported-gender-column',
    help='Column where the reported gender is held',
)
@click.option(
    '--karyotype-column',
    help='Column where the karyotype is held',
)
@click.option(
    '--reads-column',
    help='Column where the reads information is held, comma-separated if multiple',
)
@click.option(
    '--gvcf-column',
    help='Column where the reads information is held, comma-separated if multiple',
)
@click.option(
    '--participant-meta-field',
    multiple=True,
    help='Single argument, key to pull out of row to put in participant.meta',
)
@click.option(
    '--participant-meta-field-map',
    nargs=2,
    multiple=True,
    help='Two arguments per listing, eg: --participant-meta-field-map "name-in-manifest" "name-in-participant.meta"',
)
@click.option(
    '--sample-meta-field',
    multiple=True,
    help='Single argument, key to pull out of row to put in sample.meta',
)
@click.option(
    '--sample-meta-field-map',
    nargs=2,
    multiple=True,
    help='Two arguments per listing, eg: --sample-meta-field-map "name-in-manifest" "name-in-sample.meta"',
)
@click.option(
    '--assay-meta-field',
    multiple=True,
    help='Single argument, key to pull out of row to put in sample.meta',
)
@click.option(
    '--assay-meta-field-map',
    nargs=2,
    multiple=True,
    help='Two arguments per listing, eg: --assay-meta-field "name-in-manifest" "name-in-assay.meta"',
)
@click.option(
    '--qc-meta-field-map',
    nargs=2,
    multiple=True,
    help='Two arguments per listing, eg: --qc-meta-field "name-in-manifest" "name-in-analysis.meta"',
)
@click.option('--default-sample-type', default='blood')
@click.option(
    '--confirm', is_flag=True, help='Confirm with user input before updating server'
)
@click.argument('manifests', nargs=-1)
@run_as_sync
async def main(
    manifests,
    project,
    search_path: list[str],
    sample_primary_eid_column: str,
    participant_column: str | None = None,
    reported_sex_column: str | None = None,
    reported_gender_column: str | None = None,
    karyotype_column: str | None = None,
    participant_meta_field: list[str] | None = None,
    participant_meta_field_map: list[tuple[str, str]] | None = None,
    sample_meta_field: list[str] | None = None,
    sample_meta_field_map: list[tuple[str, str]] | None = None,
    assay_meta_field: list[str] | None = None,
    assay_meta_field_map: list[tuple[str, str]] | None = None,
    qc_meta_field_map: list[tuple[str, str]] = None,
    reads_column: str | None = None,
    gvcf_column: str | None = None,
    default_sample_type: str | None = None,
    confirm=False,
):
    """Run script from CLI arguments"""
    if not manifests:
        raise ValueError('Expected at least 1 manifest')

    extra_search_paths = [m for m in manifests if m.startswith('gs://')]
    if extra_search_paths:
        search_path = list(set(search_path).union(set(extra_search_paths)))

    participant_meta_map: dict[Any, Any] = {}
    sample_meta_map: dict[Any, Any] = {}
    assay_meta_map: dict[Any, Any] = {}

    qc_meta_map = dict(qc_meta_field_map or {})
    if participant_meta_field_map:
        participant_meta_map.update(dict(participant_meta_map))
    if participant_meta_field:
        participant_meta_map.update({k: k for k in participant_meta_field})
    if sample_meta_field_map:
        sample_meta_map.update(dict(sample_meta_field_map))
    if sample_meta_field:
        sample_meta_map.update({k: k for k in sample_meta_field})
    if assay_meta_field_map:
        assay_meta_map.update(dict(assay_meta_field_map))
    if assay_meta_field:
        assay_meta_map.update({k: k for k in assay_meta_field})

    parser = GenericMetadataParser(
        project=project,
        sample_primary_eid_column=sample_primary_eid_column,
        participant_primary_eid_column=participant_column,
        reported_sex_column=reported_sex_column,
        reported_gender_column=reported_gender_column,
        karyotype_column=karyotype_column,
        participant_meta_map=participant_meta_map,
        sample_meta_map=sample_meta_map,
        assay_meta_map=assay_meta_map,
        qc_meta_map=qc_meta_map,
        reads_column=reads_column,
        gvcf_column=gvcf_column,
        default_sample_type=default_sample_type,
        default_sequencing=DefaultSequencing(
            seq_type='genome', technology='short-read', platform='illumina'
        ),
        search_locations=search_path,
    )
    for manifest in manifests:
        logger.info(f'Importing {manifest}')

        await parser.from_manifest_path(manifest=manifest, confirm=confirm)


if __name__ == '__main__':
    # pylint: disable=no-value-for-parameter
    main()
